' -------------------------------------------------------------------------------
' --                                                                           --
' --                            Visual Basic Script                            --
' --                        Software Updater / Installer                       --
' --                                                                           --
' --                Copyright (c) 2019-2020 JEOL RESONANCE Inc.                --
' --                             All Rights Reserved                           --
' --                                                                           --
' -------------------------------------------------------------------------------


' configuration -----------------------------------------------------------------
INSTALLER_TITLE = "Machine Time Calculator v1.5"
   FEATURE_NAME = "Machine Time Calculator"
 TARGET_VERSION = ""

' global variables --------------------------------------------------------------
Set objFileSystem = CreateObject("Scripting.FileSystemObject")
Set      objShell = CreateObject("WScript.Shell")
         OS_SLASH = Chr(92)
       ForReading = 1
       ForWriting = 2


' installer specific functions --------------------------------------------------
Sub CheckInstalled(objInfo, strPrefix)
  ' check target software is installed/updated or not
  
  ' check user_config
  strFile = objInfo(strPrefix & "user_config")
  objInfo(strPrefix & "installed") = HasLine(strFile, FEATURE_NAME)
  objInfo(strPrefix & "is old")    = HasLine(strFile, FEATURE_NAME) And _
                                    (HasLine(strFile, INSTALLER_TITLE) = False)
End Sub

Sub DoInstall(objInfo)
  ' do installation
  
  ' create temporary folder
  strTempFolder = CreateTempFolder(objInfo)
  
  ' extract Percival files
  strPercivalFile1 = ExtractFile("machine_time_calculator.p", strTempFolder, get_machine_time_calculator_p())
  strPercivalFile2 = ExtractFile("calendar_widget.t",         strTempFolder,         get_calendar_widget_t())
  
  ' compile Percival files
  ExecuteCompiler objInfo, strPercivalFile1
  
  ' delete temporary folder
  DeleteTempFolder strTempFolder
  
  ' add tool menu to user_config
  AddMenu objInfo, INSTALLER_TITLE, "machine_time_calculator"
End Sub

Sub DoUninstall(objInfo)
  ' do uninstallation
  
  ' delete local Percival library
  strPrefix = objInfo("index") & ":"
  strFolder = objInfo(strPrefix & "library")
  strFile   = strFolder & OS_SLASH & "machine_time_calculator.jpf"
  DeleteFile strFile
  
  ' delete remaining files
  If CountFiles(strFolder, ".jpf") = 1 Then
    For Each strFile In Array("local_type_library.jpf")
      DeleteFile strFolder & OS_SLASH & strFile
    Next
  End If
  
  ' delete tool menu from user_config
  DelMenu objInfo, FEATURE_NAME
End Sub


' common functions --------------------------------------------------------------
Sub Append(strData, objValue, strDelimiter)
  ' append string with strDelimiter
  If strData <> "" Then strData = strData & strDelimiter
  strData = strData & CStr(objValue)
End Sub

Function CompareVersion(strLeftVer, strSign, strRightVer)
  ' compare version number
  arLeftVer  = Split(strLeftVer,  ".")
  arRightVer = Split(strRightVer, ".")
  For i = 0 to UBound(arLeftVer)
    nLeft  = CInt( arLeftVer(i))
    nRight = CInt(arRightVer(i))
    If (strSign = "=" And Not(nLeft = nRight)) Or _
       (strSign = ">" And    (nLeft < nRight)) Or _
       (strSign = "<" And    (nLeft > nRight)) Then
      CompareVersion = False
      Exit Function
    ElseIf (strSign = ">" And (nLeft > nRight)) Or _
           (strSign = "<" And (nLeft < nRight)) Then
      CompareVersion = True
      Exit Function
    End If
  Next
  CompareVersion = (strSign = "=")
End Function

Function GetUniquePath(strBasePath, strExtension)
  ' get unique file/folder path
  int nCount = 0
  Do
    ' get new path
    strPath = strBasePath
    If nCount > 0 Then strPath = strPath & "-" & nCount
    nCount = nCount + 1
    strPath = strPath & strExtension
    
    ' check existence
    If (strExtension  = "" And Not objFileSystem.FolderExists(strPath)) Or _
       (strExtension <> "" And Not objFileSystem.FileExists(  strPath)) Then
      GetUniquePath = strPath
      Exit Function
    End If
  Loop
End Function

Function HasLine(strFile, strTarget)
  ' check strFile has strTarget
  HasLine = False
  If objFileSystem.FileExists(strFile) Then
    Set objTextFile = objFileSystem.OpenTextFile(strFile, ForReading)
    Do Until objTextFile.AtEndOfStream
      strLine = objTextFile.ReadLine()
      If InStr(strLine, strTarget) Then
        ' strFile has strTarget
        HasLine = True
        Exit Do
      End If
    Loop
    objTextFile.Close
  End If
End Function

Function GetFolderPath(strFolderName)
  ' get folder path
  Set colProcess = objShell.Environment("PROCESS")
  strFolderPath  = ""
  
  Select Case strFolderName
    Case "ProgramFiles(32bit)"
      strFolderPath = GetFolderPath("ProgramFiles(X86)")
      If strFolderPath = "" Then
        strFolderPath = GetFolderPath("ProgramFiles")
      End If
    
    Case "ProgramFiles(64bit)"
      strFolderPath = GetFolderPath("ProgramFiles(X86)")
      If strFolderPath <> "" Then
        strFolderPath = GetFolderPath("ProgramFiles")
      End If
    
    Case Else
      strFolderPath = colProcess(strFolderName)
  End Select
  
  GetFolderPath = strFolderPath
End Function

Function CountFiles(strFolder, strExt)
  ' count files in strFolder which match strFilter
  nCount = 0
  Set objFolder = objFileSystem.GetFolder(strFolder)
  For Each objFile In objFolder.Files
    If Right(objFile.Name, 4) = strExt Then
      nCount = nCount + 1
    End If
  Next
  CountFiles = nCount
End Function

Function CreateTempFolder(objInfo)
  ' create temporary folder
  strFolder = GetUniquePath(objInfo("temp folder") & OS_SLASH & "delta_updater-temp_folder", "")
  objFileSystem.CreateFolder strFolder
  CreateTempFolder = strFolder
End Function

Sub DeleteTempFolder(strFolder)
  ' delete temporary folder
  objFileSystem.DeleteFolder strFolder, True
End Sub

Sub DeleteFile(strFile)
  On Error Resume Next
  objFileSystem.DeleteFile strFile, True
End Sub

Function ExtractFile(strFile, strFolder, strText)
  ' extract strFile from this script file into strFolder
  
  ' initialize
  strDstFile = strFolder & OS_SLASH & strFile
  
  ' write strDstFile
  Set objDstFile = objFileSystem.OpenTextFile(strDstFile, ForWriting, True)
  objDstFile.Write strText
  objDstFile.Close
  
  ' return strDstFile
  ExtractFile = strDstFile
End Function

Sub ExecuteCompiler(objInfo, strArguments)
  ' compile Percival file
  strPrefix   = objInfo("index") & ":"
  strCompiler = objInfo(strPrefix & "compiler")
  strCommand  = """" & strCompiler & """ " & strArguments
  objShell.Run strCommand, 0, True
End Sub

Sub ShowDialog(objInfo)
  ' show progress dialog
  
  ' write vbs file
  strText = "MsgBox ""Now installing " & FEATURE_NAME & " ..."", vbOKonly, """ & INSTALLER_TITLE & """"
  strFile = GetUniquePath(objInfo("temp folder") & OS_SLASH & "delta_updater-show_dialog", ".vbs")
  Set objFile = objFileSystem.OpenTextFile(strFile, ForWriting, True)
  objInfo("vbs file") = strFile
  objFile.WriteLine strText
  objFile.Close
  
  ' execute vbs file
  Set objExec = objShell.Exec("wscript.exe " & strFile)
  Set objInfo("exec object") = objExec
End Sub

Sub CloseDialog(objInfo)
  ' close progress dialog
  Set objExec = objInfo("exec object")
  strFile     = objInfo("vbs file")
  
  ' terminate the dialog
  objExec.Terminate
  
  ' delete vbs file
  objFileSystem.DeleteFile strFile, True
End Sub

Sub AddMenu(objInfo, strTitle, strProgram)
  ' add tool menu to user_config
  
  ' initialize
  strPrefix = objInfo("index") & ":"
  strFile   = objInfo(strPrefix & "user_config")
  strMenu   = strTitle & "|" & strProgram
  
  ' read existing user_config
  strText = ""
  If objFileSystem.FileExists(strFile) Then
    Set objFile = objFileSystem.OpenTextFile(strFile, ForReading)
    Do Until objFile.AtEndOfStream
      strLine = objFile.ReadLine
      strText = strText & strLine & vbCrLf
    Loop
    objFile.Close
  End If
  
  ' add tool menu
  strText = strText & strMenu & vbCrLf
  
  ' write user_config
  Set objFile = objFileSystem.OpenTextFile(strFile, ForWriting, True)
  objFile.Write strText
  objFile.Close
End Sub

Sub DelMenu(objInfo, strTitle)
  ' delete tool menu from user_config
  
  ' initialize
  strPrefix = objInfo("index") & ":"
  strFile   = objInfo(strPrefix & "user_config")
  
  ' read existing user_config
  strText = ""
  Set objFile = objFileSystem.OpenTextFile(strFile, ForReading)
  Do Until objFile.AtEndOfStream
    strLine = objFile.ReadLine
    If InStr(strLine, strTitle) <> 1 Then strText = strText & strLine & vbCrLf
  Loop
  objFile.Close
  
  ' check strText
  If strText = "" Then
    ' delete user_config
    objFileSystem.DeleteFile strFile, True
  Else
    ' write user_config
    Set objFile = objFileSystem.OpenTextFile(strFile, ForWriting, True)
    objFile.Write strText
    objFile.Close
  End If
End Sub


' functions ---------------------------------------------------------------------
Function IsDeltaRunning
  ' check delta.exe is running
  Set objWMIService  = GetObject("Winmgmts:")
  Set colProcessList = objWMIService.ExecQuery("SELECT * FROM Win32_Process WHERE Name='delta.exe'")
  bRunning = colProcessList.Count <> 0
  
  If bRunning Then
    ' delta.exe is running
    MsgBox "ERROR : Delta is running" & vbCrLf &_
           ""                         & vbCrLf &_
           "Please quit Delta and try again."  ,_
           vbCritical, INSTALLER_TITLE
  End If
  
  IsDeltaRunning = bRunning
End Function

Function FindDelta(objInfo)
  ' find Delta to install/uninstall the feature
  
  ' get folder path
  strTempFolder   = GetFolderPath("TEMP")
  strLocalFolder  = GetFolderPath("LOCALAPPDATA") & OS_SLASH & "JEOL"
  strProgFolder32 = GetFolderPath("ProgramFiles(32bit)")
  strProgFolder64 = GetFolderPath("ProgramFiles(64bit)")
  strJeolFolder32 = strProgFolder32 & OS_SLASH & "JEOL"
  strJeolFolder64 = strProgFolder64 & OS_SLASH & "JEOL"
  objInfo("temp folder") = strTempFolder
  
  ' find global JEOL folder
  If Not objFileSystem.FolderExists(strJeolFolder32) And _
     Not objFileSystem.FolderExists(strJeolFolder64) Then
    ' JEOL folder does not exist
    MsgBox "ERROR : No Delta is installed"               & vbCrLf &_
           ""                                            & vbCrLf &_
           "Please install Delta software at first and try again.",_
           vbCritical, INSTALLER_TITLE
    FindDelta = False
    Exit Function
  End If
  
  ' find Delta
  nCount      = 0
  strWithCfr  = ""
  strNoSource = ""
  strInvalids = ""
  For i = 0 To 1
    ' initialize
    If i = 0 Then
      ' find 32bit Delta
      strProgFolder = strProgFolder32
      strJeolFolder = strJeolFolder32
    Else
      ' find 64bit Delta
      strProgFolder = strProgFolder64
      strJeolFolder = strJeolFolder64
    End If
    
    ' check strJeolFolder
    If objFileSystem.FolderExists(strJeolFolder) Then
      Set colSubFolders = objFileSystem.GetFolder(strJeolFolder).SubFolders
      For Each strPath In colSubFolders
        strFolder = objFileSystem.GetFileName(strPath)
        If InStr(strFolder, "Delta ") = 1 And InStr(strFolder, ".app") = 12 Then
          ' store version
          ' Delta 5.0.5.app => 5.0.5
          nCount = nCount + 1
          strPrefix = nCount & ":"
          objInfo(strPrefix & "version")      = Mid(strFolder, 7, 3)
          objInfo(strPrefix & "full version") = Mid(strFolder, 7, 5)
          
          ' store name
          ' Delta v5.0.5 => name
          ' Delta 5.0    => folder name
          objInfo(strPrefix & "name")        = "Delta v" & objInfo(strPrefix & "full version")
          objInfo(strPrefix & "folder name") = "Delta "  & objInfo(strPrefix & "version")
          strItemName = "* " & objInfo(strPrefix & "name")
          
          ' check global source folder
          strFolder = strPath & OS_SLASH & "Contents\Resources\global\source"
          objInfo(strPrefix & "has source") = objFileSystem.GetFolder(strFolder).Files.Count > 0
          If Not objInfo(strPrefix & "has source") Then Append strNoSource, strItemName, vbCrLf
          
          ' store compiler path
          strFile = strPath & OS_SLASH & "Contents\WinNT\jdc.exe"
          objInfo(strPrefix & "compiler") = strFile
          
          ' store local library folder path
          strLocalDeltaFolder = strLocalFolder & OS_SLASH & objInfo(strPrefix & "folder name")
          strFolder = strLocalDeltaFolder & OS_SLASH & "library"
          objInfo(strPrefix & "library") = strFolder
          
          ' store user_config file path
          strFolder = strLocalDeltaFolder & OS_SLASH & "configuration"
          strFile   = strFolder & OS_SLASH & "\user_config"
          objInfo(strPrefix & "user_config") = strFile
          
          ' check installed
          CheckInstalled objInfo, strPrefix
          
          ' check 21CFR11
          strLine       = "21CFR11 = TRUE"
          strLocalFile  = strLocalDeltaFolder & OS_SLASH & "system.cfg"
          strGlobalFile = strProgFolder & OS_SLASH & "Common Files\JEOL"& OS_SLASH & objInfo(strPrefix & "folder name") & OS_SLASH & "system.cfg"
          objInfo(strPrefix & "21cfr11") = HasLine(strLocalFile, strLine) Or HasLine(strGlobalFile, strLine)
          If objInfo(strPrefix & "21cfr11") Then Append strWithCfr, strItemName, vbCrLf
          
          ' update strInvalids
          If Not objInfo(strPrefix & "has source") Or objInfo(strPrefix & "21cfr11") Then Append strInvalids, strItemName, vbCrLf
        End If
      Next
    End If
  Next
  
  ' find installable/uninstallable Delta
  strInstalled     = ""
  strUninstalled   = ""
  For i = 1 To nCount
    ' check installable
    strPrefix = i & ":"
    If (objInfo(strPrefix & "full version") = TARGET_VERSION Or TARGET_VERSION = "") And _
        objInfo(strPrefix & "has source") And _
        objInfo(strPrefix & "21cfr11") = False Then
      ' check installed
      If objInfo(strPrefix & "installed") Then
        Append strInstalled, i, ","
      End If
      If Not objInfo(strPrefix & "installed") Or objInfo(strPrefix & "is old") Then
        Append strUninstalled, i, ","
      End If
    End If
  Next
  objInfo("installed")   = strInstalled
  objInfo("uninstalled") = strUninstalled
  
  ' show messages
  If strWithCfr <> "" Then
    MsgBox "WARNING : [21CFR11 Feature Enabled] is ON"            & vbCrLf &_
           ""                                                     & vbCrLf &_
           "[21CFR11 Feature Enabled] is ON in following Delta."  & vbCrLf &_
           strWithCfr                                             & vbCrLf &_
           FEATURE_NAME & " can not be installed"                 & vbCrLf &_
           "with this feature. Please turn off [Data Integrity]-" & vbCrLf &_
           "[21CFR11 Feature Enabled] in Delta preferences and try again.", _
           vbExclamation, INSTALLER_TITLE
  End If
  If strNoSource <> "" Then
    MsgBox "WARNING : No Percival source files" & vbCrLf &_
           ""                                                       & vbCrLf &_
           "No Percival source files in following Delta."           & vbCrLf &_
           strNoSource                                              & vbCrLf &_
           FEATURE_NAME & " can not be installed"                   & vbCrLf &_
           "without Percival source files. Please re-install Delta" & vbCrLf &_
           "software with percival option and try again."                    ,_
           vbExclamation, INSTALLER_TITLE
  End If
  
  ' check ready to install/uninstall the feature
  If strUninstalled <> "" Or strInstalled <> "" Then
    ' ready to install/uninstall the feature
    FindDelta = True
  ElseIf strInvalids <> "" Then
    ' no Delta is appropriate to install/uninstall the feature
    MsgBox "ERROR : No Delta is installable"                      & vbCrLf &_
           ""                                                     & vbCrLf &_
           "Following Delta is detected, but no Delta is ready"   & vbCrLf &_
           "for the installation of " & FEATURE_NAME & "."        & vbCrLf &_
           strInvalids                                            & vbCrLf &_
           "Please solve previous warning message and try again."          ,_
           vbCritical, INSTALLER_TITLE
    FindDelta = False
  ElseIf TARGET_VERSION <> "" Then
    ' TARGET_VERSION Delta is found
    MsgBox "ERROR : Delta v" & TARGET_VERSION & " is not installed"     & vbCrLf &_
           ""                                                           & vbCrLf &_
           "Please install Delta v" & TARGET_VERSION & " at first and try again.",_
           vbCritical, INSTALLER_TITLE
    FindDelta = False
  Else
    ' no Delta is found
    MsgBox "ERROR : no Delta is not installed"  & vbCrLf &_
           ""                                   & vbCrLf &_
           "Please install Delta at first and try again.",_
           vbCritical, INSTALLER_TITLE
    FindDelta = False
  End If
End Function

Function GetAction(objInfo)
  ' get action to do
  
  ' define menu
  nAction = 0
  strMenu = "Please select menu item." & VbCrLf
  
  ' define install menu items
  arIndex = Split(objInfo("uninstalled"), ",")
  For i = 0 To UBound(arIndex)
    nAction = nAction + 1
    strPrefix = arIndex(i) & ":"
    strActionPrefix = nAction & ":"
    objInfo(strActionPrefix & "action") = "install"
    objInfo(strActionPrefix & "index")  = arIndex(i)
    strMenu = strMenu & vbCrLf & nAction & ". Install to " & objInfo(strPrefix & "name")
  Next
  
  ' define uninstall menu items
  arIndex = Split(objInfo("installed"), ",")
  For i = 0 To UBound(arIndex)
    nAction = nAction + 1
    strPrefix = arIndex(i) & ":"
    strActionPrefix = nAction & ":"
    objInfo(strActionPrefix & "action") = "uninstall"
    objInfo(strActionPrefix & "index")  = arIndex(i)
    strMenu = strMenu & vbCrLf & nAction & ". Uninstall from " & objInfo(strPrefix & "name")
  Next
  
  ' add other menu item
  strMenu = strMenu   & vbCrLf & VbCrLf &_
            "q. Quit" & VbCrLf & VbCrLf &_
            "Enter choice:"
  
  ' show menu and get a choice
  Do
    ' show menu
    strChoice = LCase(InputBox(strMenu, INSTALLER_TITLE))
    
    ' check the choice
    If strChoice = "q" Or strChoice = "" Then
      ' quit the installer
      GetAction = False
      Exit Function
    ElseIf IsNumeric(strChoice) Then
      ' check the choice
      nChoice = CInt(strChoice)
      If nChoice > 0 And nChoice < nAction + 1 Then
        ' store the choice
        objInfo("action") = CStr(nChoice)
        GetAction = True
        Exit Function
      End If
    End If
  Loop
End Function

Sub Install(objInfo)
  ' install the feature
  
  ' show progress dialog
  ShowDialog objInfo
  
  ' uninstall previously installed feature
  strPrefix = objInfo("index") & ":"
  If objInfo(strPrefix & "is old") Then
    DoUninstall objInfo
  End If
  
  ' do installation
  DoInstall objInfo
  
  ' close progress dialog
  CloseDialog objInfo
  
  ' show message
  strPrefix = objInfo("index") & ":"
  MsgBox "Install Succeeded"                          & vbCrLf &_
         ""                                           & vbCrLf &_
         "Installed " & FEATURE_NAME                  & vbCrLf &_
         "to " & objInfo(strPrefix & "name") & " successfully.",_
         vbInformation, INSTALLER_TITLE
End Sub

Sub Uninstall(objInfo)
  ' uninstall the feature
  
  ' do uninstallation
  DoUninstall objInfo
  
  ' show message
  strPrefix = objInfo("index") & ":"
  MsgBox "Uninstall Succeeded"                          & vbCrLf &_
         ""                                             & vbCrLf &_
         "Uninstalled " & FEATURE_NAME                  & vbCrLf &_
         "from " & objInfo(strPrefix & "name") & " successfully.",_
         vbInformation, INSTALLER_TITLE
End Sub

Sub DoAction(objInfo)
  ' do selected action
  
  ' get selected action information
  strActionPrefix = objInfo("action") & ":"
  strAction = objInfo(strActionPrefix & "action")
  strIndex  = objInfo(strActionPrefix & "index")
  objInfo("index") = strIndex
  
  ' do action
  If strAction =   "install" Then   Install(objInfo)
  If strAction = "uninstall" Then Uninstall(objInfo)
End Sub


' main procedure ----------------------------------------------------------------
' check delta.exe is running
If Not IsDeltaRunning Then
  ' find Delta to install/uninstall the feature
  Set objInfo = CreateObject("Scripting.Dictionary")
  If FindDelta(objInfo) Then
    ' get action to do
    If GetAction(objInfo) Then
      ' do selected action
      DoAction objInfo
    End If
  End If
End If

' machine_time_calculator.p -----------------------------------------------------
Function get_machine_time_calculator_p()
  get_machine_time_calculator_p = _
"-------------------------------------------------------------------------------" & vbLf & _
"--                                                                           --" & vbLf & _
"--                           Percival Source Code                            --" & vbLf & _
"--                Delta NMR Processing and Control Interface                 --" & vbLf & _
"--                                                                           --" & vbLf & _
"--                Copyright (c) 2017-2020 JEOL RESONANCE Inc.                --" & vbLf & _
"--                           All Rights Reserved                             --" & vbLf & _
"--                                                                           --" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"program MACHINE_TIME_CALCULATOR is" & vbLf & _
"" & vbLf & _
"    include ""authorization"";" & vbLf & _
"    include ""auto_open"";" & vbLf & _
"    include ""calendar_widget"";" & vbLf & _
"    include ""callsend"";" & vbLf & _
"    include ""color_schemes"";" & vbLf & _
"    include ""common"";" & vbLf & _
"    include ""data_server"";" & vbLf & _
"    include ""dates"";" & vbLf & _
"    include ""file_system"";" & vbLf & _
"    include ""intents"";" & vbLf & _
"    include ""locale"";" & vbLf & _
"    include ""message_support"";" & vbLf & _
"    include ""preference"";" & vbLf & _
"    include ""strings"";" & vbLf & _
"    include ""variables"";" & vbLf & _
"" & vbLf & _
"    enumeration { MSG_INFO, MSG_WARNING, MSG_ERROR };" & vbLf & _
"" & vbLf & _
"    function SHOW_MESSAGE( msg : WILDCARD, msg_type : NUMERIC, option : STRING := """" );" & vbLf & _
"" & vbLf & _
"    template PROJECT_SUMMARY_FORMAT is { project_name, op_time, op_days, op_count, exp_count };" & vbLf & _
"    template  FOLDER_SUMMARY_FORMAT is { folder, op_time, op_days, op_count, exp_count };" & vbLf & _
"    template    USER_SUMMARY_FORMAT is { username, owner_time, owner_days, owner_count, op_count, exp_count };" & vbLf & _
"    template         PROJECT_FORMAT is { op_start, op_stop, duration, days, username, job_name, exp_file, project_name, folder, data_file };" & vbLf & _
"    template          FOLDER_FORMAT is { op_start, op_stop, duration, days, username, job_name, exp_file, project_name, folder, data_file };" & vbLf & _
"    template            USER_FORMAT is { owner_start, owner_stop, duration, days, job_count, exp_count, job_name, exp_file, project_name, folder, data_file, job_result, first_fatal };" & vbLf & _
"    template      PROJECT_TABLE_ROW is { project_name, op_time, op_count, exp_count };" & vbLf & _
"    template       FOLDER_TABLE_ROW is { folder, op_time, op_count, exp_count };" & vbLf & _
"    template         USER_TABLE_ROW is { username, owner_time, owner_count, op_count, exp_count };" & vbLf & _
"    template          OWNER_SECTION is { start, stop, username, op_sections };" & vbLf & _
"    template            EXP_SECTION is { start, stop, exp_file, folder, data_id, data_file, for_shimming, first_fatal, last_fatal };" & vbLf & _
"    template             OP_SECTION is { start, stop, job_name, job_result, project_id, project_name, first_fatal, last_fatal, did_calibration, did_shimming, exp_sections };" & vbLf & _
"    template           PROJECT_DATA is { project_name, op_time, exp_count, ops };" & vbLf & _
"    template            FOLDER_DATA is { folder, op_time, exp_count, ops };" & vbLf & _
"    template             OWNER_DATA is { start, stop, ops };" & vbLf & _
"    template             QUERY_DATA is { data_id, data_file, exp_file, project_id, project_name, folder };" & vbLf & _
"    template              USER_DATA is { username, owner_time, op_count, exp_count, owners };" & vbLf & _
"    template               EXP_DATA is { exp_file, folder, data_file };" & vbLf & _
"    template                OP_DATA is { start, stop, username, job_name, job_result, first_fatal, project_name, exps };" & vbLf & _
"" & vbLf & _
"    var" & vbLf & _
"        COLOR_ATTENTION    : constant SET     := {255,250,191}," & vbLf & _
"        COLOR_ERROR        : constant SET     := {255,154,154}," & vbLf & _
"        REG_CURR_MSG       : constant NUMERIC := 1," & vbLf & _
"        MAX_LOG_LINES      : constant NUMERIC := 100," & vbLf & _
"        DATE_LENGTH_JPN    : constant NUMERIC := 10," & vbLf & _
"        DATE_LENGTH_ENG    : constant NUMERIC := 11," & vbLf & _
"        FLAG_ABORT_ENABLED : constant NUMERIC := 2**0," & vbLf & _
"        FLAG_ABORTED       : constant NUMERIC := 2**1," & vbLf & _
"        LBL_SPEC           : constant STRING  := ""Spectrometer""," & vbLf & _
"        LBL_USER           : constant STRING  := ""Username""," & vbLf & _
"        LBL_PASS           : constant STRING  := ""Password""," & vbLf & _
"        LBL_START          : constant STRING  := ""Start Date""," & vbLf & _
"        LBL_STOP           : constant STRING  := ""Stop Date""," & vbLf & _
"        LBL_ABORT          : constant STRING  := ""Abort""," & vbLf & _
"        LBL_TITLE          : constant STRING  := ""Machine Time Calculator v1.5""," & vbLf & _
"        LBL_TARGET         : constant STRING  := ""Target Field""," & vbLf & _
"        LBL_CONNECT        : constant STRING  := ""Connect""," & vbLf & _
"        LBL_ABORTING       : constant STRING  := ""Aborting ...""," & vbLf & _
"        LBL_CALCULATE      : constant STRING  := ""Calculate""," & vbLf & _
"        LBL_DOUBLE_MINUS   : constant STRING  := ""&double_minus;""," & vbLf & _
"        INFO_ABORT         : constant STRING  := ""Click this button to abort current process.""," & vbLf & _
"        INFO_CONNECT       : constant STRING  := ""Click this button to start connecting to the spectrometer&#xA;and download log files for machine time calculation.""," & vbLf & _
"        INFO_CALCULATE     : constant STRING  := ""Click this button to start machine time calculation.""," & vbLf & _
"        BASENAMES          : constant SET     := {""machine"",""queue_history""}," & vbLf & _
"        TAB_NAMES          : constant SET     := {""Username"",""Project Name"",""Folder Name""}," & vbLf & _
"" & vbLf & _
"        conf            : SET," & vbLf & _
"        logs            : SET," & vbLf & _
"        cal_pos         : SET," & vbLf & _
"        temp_files      : SET," & vbLf & _
"        data_by_user    : SET," & vbLf & _
"        data_by_proj    : SET," & vbLf & _
"        data_by_folder  : SET," & vbLf & _
"        table_by_user   : SET," & vbLf & _
"        table_by_proj   : SET," & vbLf & _
"        table_by_folder : SET," & vbLf & _
"        owner_sections  : SET," & vbLf & _
"        def             : STRING," & vbLf & _
"        cal_id          : STRING," & vbLf & _
"        cal_target      : STRING," & vbLf & _
"        stop_date       : STRING," & vbLf & _
"        start_date      : STRING," & vbLf & _
"        target_field    : STRING," & vbLf & _
"        abort_msg       : STRING," & vbLf & _
"        abort_action    : STRING," & vbLf & _
"        next_action     : STRING," & vbLf & _
"        is_jpn          : BOOLEAN," & vbLf & _
"        shutdown        : BOOLEAN," & vbLf & _
"        use_folder      : BOOLEAN," & vbLf & _
"        folder_used     : BOOLEAN," & vbLf & _
"        nc              : NETCONTEXT;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function LOG_MESSAGE( msg : STRING, msg_type : NUMERIC := NULL ) is" & vbLf & _
"        var" & vbLf & _
"            SPACES : constant STRING := ""  ""," & vbLf & _
"" & vbLf & _
"            color : SET," & vbLf & _
"            time  : STRING;" & vbLf & _
"" & vbLf & _
"        case msg_type is" & vbLf & _
"            when MSG_INFO    => color := {0,139,0};" & vbLf & _
"            when MSG_WARNING => color := {220,70,0};" & vbLf & _
"            when MSG_ERROR   => color := {165,42,42};" & vbLf & _
"        end case;" & vbLf & _
"" & vbLf & _
"        time := remove( date( seconds ), 1, DATE_LENGTH_ENG + 1 );" & vbLf & _
"        sw {{def,""lb_log""}, {WM~ADD_VALUE, time & SPACES & msg}, (color = {}) ? {} :" & vbLf & _
"                            {WM~SET_SELECTED, 0, color}};" & vbLf & _
"    end LOG_MESSAGE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function FORMAT_TIME( time : WILDCARD ) return STRING is" & vbLf & _
"        var" & vbLf & _
"            ONE_HOUR   : constant NUMERIC := 3600," & vbLf & _
"            ONE_MINUTE : constant NUMERIC := 60," & vbLf & _
"" & vbLf & _
"            secs  : UNIVERSAL," & vbLf & _
"            mins  : UNIVERSAL," & vbLf & _
"            hours : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        time  := round( time ^[] );" & vbLf & _
"        hours := (time - (time mod ONE_HOUR)) / ONE_HOUR;" & vbLf & _
"        time  := time - hours * ONE_HOUR;" & vbLf & _
"        mins  := (time - (time mod ONE_MINUTE)) / ONE_MINUTE;" & vbLf & _
"        secs  := time - mins * ONE_MINUTE;" & vbLf & _
"" & vbLf & _
"        return format( hours, ""z99"" ) & "":"" & format( mins, ""z99"" ) & "":"" & format( secs, ""z99"" );" & vbLf & _
"    end FORMAT_TIME;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function UPDATE_INDICATOR( action : WILDCARD, rate : WILDCARD := NULL, msg : STRING := NULL ) is" & vbLf & _
"        var" & vbLf & _
"            UPDATE_INTERVAL : constant UNIVERSAL := 1[s]," & vbLf & _
"" & vbLf & _
"            start : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        case action is" & vbLf & _
"            when ""show"" =>" & vbLf & _
"                sw {def, {""lb_log"",     {WM~LEFT_BOTTOM_RIGHT_TOP, NULL,14[mm],NULL,NULL}}," & vbLf & _
"                         {""sld_status"", {WM~SET_SELECTED, 0}}," & vbLf & _
"                         {""lbl_status"", {WM~SET_VALUE, """"}," & vbLf & _
"                                        {WM~SET_REGISTER, REG_CURR_MSG, """"}}," & vbLf & _
"                         {""lbl_msg"",    {WM~SET_FEATURE, WM~NO_DRAW, TRUE}}," & vbLf & _
"                         {""fr_status"",  {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}};" & vbLf & _
"" & vbLf & _
"            when ""update"" =>" & vbLf & _
"                sw {{def,""sld_status""}, {WM~SET_SELECTED, rate}};" & vbLf & _
"                return when msg = NULL;" & vbLf & _
"" & vbLf & _
"                start := seconds;" & vbLf & _
"                log_message( msg );" & vbLf & _
"                sw {{def,""lbl_status""}, {WM~SET_VALUE, msg}};" & vbLf & _
"                cw {{def,""lbl_status""}, {WM~SET_REGISTER, REG_CURR_MSG, msg}};" & vbLf & _
"" & vbLf & _
"                spawn block" & vbLf & _
"                    loop" & vbLf & _
"                        delay UPDATE_INTERVAL;" & vbLf & _
"                        exit when msg /= cw {{def,""lbl_status""}, {WM~GET_REGISTER, REG_CURR_MSG}};" & vbLf & _
"                        sw {{def,""lbl_status""}, {WM~SET_VALUE, msg & "" "" & format_time( seconds - start )}};" & vbLf & _
"                    end loop;" & vbLf & _
"                end block;" & vbLf & _
"" & vbLf & _
"            when ""hide"" =>" & vbLf & _
"                sw {def, {""lb_log"",     {WM~LEFT_BOTTOM_RIGHT_TOP, NULL,12[mm],NULL,NULL}}," & vbLf & _
"                         {""fr_status"",  {WM~SET_FEATURE, WM~NO_DRAW, TRUE}}," & vbLf & _
"                         {""lbl_status"", {WM~SET_REGISTER, REG_CURR_MSG, """"}}};" & vbLf & _
"        end case;" & vbLf & _
"    end UPDATE_INDICATOR;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function HANDLE_STATE( action : STRING, arg1 : WILDCARD := NULL, arg2 : WILDCARD := NULL ) return BOOLEAN is" & vbLf & _
"        case action is" & vbLf & _
"            when ""enable abort"" =>" & vbLf & _
"                abort_msg    := arg1;" & vbLf & _
"                abort_action := arg2;" & vbLf & _
"                update_indicator( ""show"" );" & vbLf & _
"                sw {def, {""pb_action"",     {WM~SET_VALUE, LBL_ABORT}," & vbLf & _
"                                           {WM~SET_INFO, INFO_ABORT}," & vbLf & _
"                                           {WM~SET_HIGHLIGHT, ""Sienna1""}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORT_ENABLED, TRUE}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORTED,      FALSE}}," & vbLf & _
"                         {""fr_right"",      {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""fr_left"",       {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""pb_save"",       {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""lb_log"",        {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}," & vbLf & _
"                         {""tabs"",          {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}," & vbLf & _
"                         {""tbl_data"",      {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}};" & vbLf & _
"" & vbLf & _
"            when ""disable connect"" =>" & vbLf & _
"                update_indicator( ""hide"" );" & vbLf & _
"                sw {def, {""pb_action"",     {WM~SET_VALUE, LBL_CONNECT}," & vbLf & _
"                                           {WM~SET_INFO, INFO_CONNECT}," & vbLf & _
"                                           {WM~SET_HIGHLIGHT, color_accept}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORT_ENABLED, FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORTED,       FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""fr_right"" ,     {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""fr_left"" ,      {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""pb_save"",       {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""lb_log"",        {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}," & vbLf & _
"                         {""tabs"",          {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}," & vbLf & _
"                         {""tbl_data"",      {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}};" & vbLf & _
"" & vbLf & _
"            when ""enable connect"" =>" & vbLf & _
"                update_indicator( ""hide"" );" & vbLf & _
"                sw {def, {""pb_action"",     {WM~SET_VALUE, LBL_CONNECT}," & vbLf & _
"                                           {WM~SET_INFO, INFO_CONNECT}," & vbLf & _
"                                           {WM~SET_HIGHLIGHT, color_accept}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORT_ENABLED, FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORTED,       FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""fr_left"" ,      {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""fr_right"" ,     {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""pb_save"",       {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""lb_log"",        {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}," & vbLf & _
"                         {""tabs"",          {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}," & vbLf & _
"                         {""tbl_data"",      {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}};" & vbLf & _
"" & vbLf & _
"            when ""enable calculate"" =>" & vbLf & _
"                update_indicator( ""hide"" );" & vbLf & _
"                sw {def, {""ib_start_date"", {WM~SET_VALUE, start_date}}," & vbLf & _
"                         {""ib_stop_date"",  {WM~SET_VALUE,  stop_date}}," & vbLf & _
"                         {""pb_action"",     {WM~SET_VALUE, LBL_CALCULATE}," & vbLf & _
"                                           {WM~SET_INFO, INFO_CALCULATE}," & vbLf & _
"                                           {WM~SET_HIGHLIGHT, color_accept}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORT_ENABLED, FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORTED,       FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""pb_save"",       {WM~SET_FEATURE, WM~ENABLE,  FALSE}}," & vbLf & _
"                         {""fr_right"",      {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""lb_log"",        {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}," & vbLf & _
"                         {""tabs"",          {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}," & vbLf & _
"                         {""tbl_data"",      {WM~SET_FEATURE, WM~NO_DRAW,  TRUE}}};" & vbLf & _
"" & vbLf & _
"            when ""enable save"" =>" & vbLf & _
"                update_indicator( ""hide"" );" & vbLf & _
"                sw {def, {""pb_action"",     {WM~SET_VALUE, LBL_CALCULATE}," & vbLf & _
"                                           {WM~SET_INFO, INFO_CALCULATE}," & vbLf & _
"                                           {WM~SET_HIGHLIGHT, color_accept}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORT_ENABLED, FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, FLAG_ABORTED,       FALSE}," & vbLf & _
"                                           {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""pb_save"",       {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""fr_right"",      {WM~SET_FEATURE, WM~ENABLE,   TRUE}}," & vbLf & _
"                         {""tabs"",          {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}};" & vbLf & _
"                sw {WM~SEND_ADDRESS, {def,""tabs""}, {WM~RETURN_STATE, WM~BUTTON_ON, cw {{def,""tabs""}, {WM~GET_SELECTED}}}};" & vbLf & _
"" & vbLf & _
"            when ""is aborted"" =>" & vbLf & _
"                if (cw {{def,""pb_action""}, {WM~GET_FEATURE, FLAG_ABORTED}}) /= 0 then" & vbLf & _
"                    -- aborted" & vbLf & _
"                    -- show abort message and do abort action" & vbLf & _
"                    if abort_msg /= NULL then" & vbLf & _
"                        show_message( abort_msg, MSG_WARNING, abort_action );" & vbLf & _
"                    end if;" & vbLf & _
"                    return TRUE;" & vbLf & _
"                else" & vbLf & _
"                    -- not aborted" & vbLf & _
"                    -- update current states" & vbLf & _
"                    if arg1 /= NULL then" & vbLf & _
"                        update_indicator( ""update"", arg1 );" & vbLf & _
"                    end if;" & vbLf & _
"                    if arg2 /= NULL then" & vbLf & _
"                        arg2 := seconds;" & vbLf & _
"                    end if;" & vbLf & _
"                    return FALSE;" & vbLf & _
"                end if;" & vbLf & _
"        end case;" & vbLf & _
"" & vbLf & _
"        return TRUE;" & vbLf & _
"    end HANDLE_STATE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function SHOW_MESSAGE( msg : WILDCARD, msg_type : NUMERIC, option : STRING := """" ) is" & vbLf & _
"        var" & vbLf & _
"            color : SET," & vbLf & _
"            icon  : STRING;" & vbLf & _
"" & vbLf & _
"        case msg_type is" & vbLf & _
"            when MSG_INFO =>" & vbLf & _
"                icon  := ""info_bubble_small.rgb"";" & vbLf & _
"                color := COLOR_ATTENTION;" & vbLf & _
"" & vbLf & _
"            when MSG_WARNING =>" & vbLf & _
"                icon  := ""alert_small.rgb"";" & vbLf & _
"                color := COLOR_ATTENTION;" & vbLf & _
"" & vbLf & _
"            when MSG_ERROR =>" & vbLf & _
"                icon  := ""x_red_small.rgb"";" & vbLf & _
"                color := COLOR_ERROR;" & vbLf & _
"        end case;" & vbLf & _
"" & vbLf & _
"        msg := (type msg = [STRING]) ? {msg} : msg;" & vbLf & _
"        if msg_type /= MSG_INFO then" & vbLf & _
"            log_message( msg(1), msg_type );" & vbLf & _
"        end if;" & vbLf & _
"        for i : NUMERIC is 1 to size( msg ) loop" & vbLf & _
"            msg(i) := ""  "" & msg(i);" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        sw {{def,""lbl_msg""}, {WM~SET_VALUE, msg}," & vbLf & _
"                             {WM~SET_VALUE, WM~PIXMAP, icon}," & vbLf & _
"                             {WM~SET_BACKGROUND, color}," & vbLf & _
"                             {WM~SET_FEATURE, WM~NO_DRAW, FALSE}};" & vbLf & _
"" & vbLf & _
"        handle_state( option );" & vbLf & _
"    end SHOW_MESSAGE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define CHECK_STATUS is" & vbLf & _
"        var" & vbLf & _
"            msg    : STRING," & vbLf & _
"            option : STRING," & vbLf & _
"            spec   : STRING := cw {{def,""se_spec""}, {WM~GET_SELECTED}}," & vbLf & _
"            user   : STRING := cw {{def,""ib_user""}, {WM~GET_SELECTED}};" & vbLf & _
"" & vbLf & _
"        if spec = """" then" & vbLf & _
"            if next_action /= ""connect"" then" & vbLf & _
"                show_message( ""Spectrometer was disconnected."", MSG_WARNING );" & vbLf & _
"                next_action := ""connect"";" & vbLf & _
"            end if;" & vbLf & _
"            msg    := ""Please select spectrometer."";" & vbLf & _
"            option := ""disable connect"";" & vbLf & _
"        elsif user = """" then" & vbLf & _
"            msg    := ""Please type username and password."";" & vbLf & _
"            option := ""disable connect"";" & vbLf & _
"        elsif next_action /= ""connect"" then" & vbLf & _
"            return;" & vbLf & _
"        else" & vbLf & _
"            msg    := ""Please click [Connect] button to check log files."";" & vbLf & _
"            option := ""enable connect"";" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        show_message( msg, MSG_INFO, option );" & vbLf & _
"    end CHECK_STATUS;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define UPDATE_SPECS is" & vbLf & _
"        var" & vbLf & _
"            specs : SET    := find_devices( ""JEOL Spectrometer"" )," & vbLf & _
"            spec  : STRING := string_pref(  ""MACHINE_TIME_CALCULATOR.SPECTROMETER"", """" );" & vbLf & _
"" & vbLf & _
"        specs := {""""} & (size( specs ) = 0) ? {} : sort( (explode( specs ))(2) );" & vbLf & _
"        spec  := (spec in specs) ? spec : """";" & vbLf & _
"        sw {{def,""se_spec""}, {WM~SET_VALUE, {specs, spec}}};" & vbLf & _
"        check_status;" & vbLf & _
"    end UPDATE_SPECS;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function GET_TAB_ROTATION( widgets : SET ) return SET is" & vbLf & _
"        var" & vbLf & _
"            msg : SET," & vbLf & _
"            sz  : NUMERIC := size( widgets );" & vbLf & _
"" & vbLf & _
"        for i : NUMERIC is 1 to sz loop" & vbLf & _
"            msg := msg & {{" & vbLf & _
"                widgets(i)} & ((i > 1) ? {} : {" & vbLf & _
"                    {WM~SET_FOCUS}}) & {" & vbLf & _
"                    {WM~SET_STATE, WM~TAB_KEY, widgets((i < sz) ? (i+1) : 1)}}};" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        return msg;" & vbLf & _
"    end GET_TAB_ROTATION;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define INITIALIZE is" & vbLf & _
"        var" & vbLf & _
"            LBL_WIDTH       : constant UNIVERSAL := max (cw {WM~GET_STRING_WIDTH, {LBL_CONNECT,LBL_ABORTING     }}) + 4[mm]," & vbLf & _
"            LBL_WIDTH_LEFT  : constant UNIVERSAL := max (cw {WM~GET_STRING_WIDTH, {LBL_SPEC,LBL_USER,LBL_PASS   }}) + 2[mm]," & vbLf & _
"            LBL_WIDTH_RIGHT : constant UNIVERSAL := max (cw {WM~GET_STRING_WIDTH, {LBL_START,LBL_STOP,LBL_TARGET}}) + 2[mm];" & vbLf & _
"" & vbLf & _
"        def := cw ({" & vbLf & _
"            {WM~FRAME, ""fr_main"", WM~SCREEN, ""machine_time_calculator""," & vbLf & _
"                {WM~ICON_NAME,  ""machine_time_calculator""}," & vbLf & _
"                {WM~ICON_TITLE, LBL_TITLE}," & vbLf & _
"                {WM~SET_VALUE,  LBL_TITLE}," & vbLf & _
"                {WM~SET_STATE, WM~MINIMUM, 145[mm], 100[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~TRACK_POSITION, boolean_pref( ""SYSTEM.DELTA_TOOLS.PRESERVE_STATE"", FALSE )}," & vbLf & _
"                {WM~BINDING, {WM~KEYBOARD,WM~ESC_KEY,WM~PRESS,WM~NOSHIFT,WM~NOCONTROL,WM~NOALTERATION,WM~NOMODIFIER}, WM~REPROCESS, 0," & vbLf & _
"                             {WM~WINQUIT}}," & vbLf & _
"                {WM~BINDING, {WM~WINQUIT}, WM~SEND_PERCIVAL, 0, {""quit""}}" & vbLf & _
"            }," & vbLf & _
"            {WM~FRAME, ""fr_left"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-28[mm],50[%],1[mm]}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_spec"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-9[mm],-LBL_WIDTH_LEFT,0[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_SPEC}," & vbLf & _
"                {WM~SET_INFO, ""Target spectrometer to calculate machine time""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_user"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-18[mm],-LBL_WIDTH_LEFT,9[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_USER}," & vbLf & _
"                {WM~SET_INFO, ""Username of the administrative account for the spectrometer""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_pass"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-27[mm],-LBL_WIDTH_LEFT,18[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_PASS}," & vbLf & _
"                {WM~SET_INFO, ""Password of the administrative account for the spectrometer""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~SELECT_ENUMERATION, ""se_spec"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_LEFT,-9[mm],1[mm],0[mm]}," & vbLf & _
"                {WM~SET_VALUE, {{""""}, """"}}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""spec"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~INPUT_BOX, ""ib_user"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_LEFT,-18[mm],1[mm],9[mm]}," & vbLf & _
"                {WM~SET_VALUE, string_pref( ""MACHINE_TIME_CALCULATOR.USERNAME"", """" )}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""user"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~INPUT_BOX, ""ib_pass"", ""fr_left""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_LEFT,-27[mm],1[mm],18[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~PASSWORD, TRUE}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""pass"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~FRAME, ""fr_right"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 50[%],-28[mm],0[mm],1[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~DISABLE, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_start_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-9[mm],-LBL_WIDTH_RIGHT,0[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_START}," & vbLf & _
"                {WM~SET_INFO, ""Start date for the machine time calculation""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_stop_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-18[mm],-LBL_WIDTH_RIGHT,9[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_STOP}," & vbLf & _
"                {WM~SET_INFO, ""Stop date for the machine time calculation""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_target"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-27[mm],-LBL_WIDTH_RIGHT,18[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_TARGET}," & vbLf & _
"                {WM~SET_INFO, ""Field to be referred as the target of the date range""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~CONTOUR_SYMMETRICALLY, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~INPUT_BOX, ""ib_start_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_RIGHT,-9[mm],9[mm],0[mm]}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""start date"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~INPUT_BOX, ""ib_stop_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_RIGHT,-18[mm],9[mm],9[mm]}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""stop date"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~SELECT_ENUMERATION, ""se_target"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, LBL_WIDTH_RIGHT,-27[mm],9[mm],18[mm]}," & vbLf & _
"                {WM~SET_VALUE, {{""Owner Began"",""Operation Began""}, string_pref( ""MACHINE_TIME_CALCULATOR.TARGET"", ""Owner Began"" )}}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_SELECTED}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",WM~INSERT,""target"",WM~REMAINING}}" & vbLf & _
"            }," & vbLf & _
"            {WM~TOGGLE_BUTTON, ""tb_start_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, -8[mm],-8[mm],1[mm],1[mm]}," & vbLf & _
"                {WM~SET_VALUE, WM~PIXMAP, ""calendar_small.rgb""}," & vbLf & _
"                {WM~SET_INFO, ""Toggle calendar widget for specifying start date""}," & vbLf & _
"                {WM~SET_FEATURE, WM~PERSPECTIVE, FALSE}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""calendar"",WM~INSERT,""start date"",WM~WILDCARD}}" & vbLf & _
"            }," & vbLf & _
"            {WM~TOGGLE_BUTTON, ""tb_stop_date"", ""fr_right""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, -8[mm],-17[mm],1[mm],10[mm]}," & vbLf & _
"                {WM~SET_VALUE, WM~PIXMAP, ""calendar_small.rgb""}," & vbLf & _
"                {WM~SET_INFO, ""Toggle calendar widget for specifying stop date""}," & vbLf & _
"                {WM~SET_FEATURE, WM~PERSPECTIVE, FALSE}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""calendar"",WM~INSERT,""stop date"",WM~WILDCARD}}" & vbLf & _
"            }," & vbLf & _
"            {WM~RADIO_BUTTON, ""tabs"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-39[mm],0[mm],29[mm]}," & vbLf & _
"                {WM~SET_VALUE, TAB_NAMES}," & vbLf & _
"                {WM~SET_INFO,  {""Machine times grouped by usernames""," & vbLf & _
"                                ""Machine times grouped by project names""," & vbLf & _
"                                ""Machine times grouped by folder names""}}," & vbLf & _
"                {WM~SET_FEATURE, WM~NO_DRAW, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~PERSPECTIVE, FALSE}," & vbLf & _
"                {WM~SET_STATE, WM~BUTTON_ON, index( string_pref( ""MACHINE_TIME_CALCULATOR.GROUP"", ""Username"" ), TAB_NAMES )}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON,1}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",""group"",""Username""}}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON,2}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",""group"",""Project Name""}}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON,3}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""set value"",""group"",""Folder Name""}}" & vbLf & _
"            }," & vbLf & _
"            {WM~PUSH_BUTTON, ""pb_action"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, -(LBL_WIDTH + 9[mm]),-36.5[mm],9[mm],28.5[mm]}," & vbLf & _
"                {WM~SET_VALUE, LBL_CONNECT}," & vbLf & _
"                {WM~SET_INFO, INFO_CONNECT}," & vbLf & _
"                {WM~SET_HIGHLIGHT, color_accept}," & vbLf & _
"                {WM~SET_FEATURE, WM~DISABLE, TRUE}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}, WM~NOT_SEND_PERCIVAL, FLAG_ABORT_ENABLED," & vbLf & _
"                             {""action""}}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}, WM~REPROCESS, FLAG_ABORT_ENABLED," & vbLf & _
"                             {""show aborting""}}," & vbLf & _
"                {WM~BINDING, {""show aborting""}, WM~REPROCESS, FLAG_ABORT_ENABLED," & vbLf & _
"                             {WM~SET_VALUE, ""Aborting ...""}}," & vbLf & _
"                {WM~BINDING, {""show aborting""}, WM~REPROCESS, FLAG_ABORT_ENABLED," & vbLf & _
"                             {WM~SET_FEATURE, WM~DISABLE, TRUE}}," & vbLf & _
"                {WM~BINDING, {""show aborting""}, WM~SET_FLAG, FLAG_ABORTED}," & vbLf & _
"                {WM~BINDING, {""show aborting""}, WM~CLEAR_FLAG, FLAG_ABORT_ENABLED}" & vbLf & _
"            }," & vbLf & _
"            {WM~PUSH_BUTTON, ""pb_save"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, -8[mm],-36.5[mm],1[mm],28.5[mm]}," & vbLf & _
"                {WM~SET_VALUE, WM~PIXMAP, ""floppy_savetext_small.rgb""}," & vbLf & _
"                {WM~SET_INFO, ""Click this button to save current machine time data into CSV file.&#xA;You can narrow down detailed part in CSV file&#xA;by selecting target rows in the table.""}," & vbLf & _
"                {WM~SET_FEATURE, WM~ENABLE, FALSE}," & vbLf & _
"                {WM~BINDING, {WM~RETURN_STATE}, WM~SEND_PERCIVAL, 0," & vbLf & _
"                             {""save""}}" & vbLf & _
"            }," & vbLf & _
"            {WM~LIST_BOX, ""lb_log"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 1[mm],12[mm],1[mm],38[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~BORDER, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~GRID, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~UNDERLINE, TRUE}," & vbLf & _
"                {WM~SET_STATE, WM~MAX_LINES, MAX_LOG_LINES}" & vbLf & _
"            }," & vbLf & _
"            {WM~TABLE, ""tbl_data"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 1[mm],12[mm],1[mm],39.5[mm]}," & vbLf & _
"                {WM~SET_VALUE, WM~SET_MIN_COLUMN_WIDTH, {37[mm],NULL,NULL,NULL,NULL,NULL,NULL,NULL}}," & vbLf & _
"                {WM~SET_FEATURE, WM~GRID, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~NO_DRAW, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~SEQUENCE, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~SEQUENCE_MARKER, TRUE}," & vbLf & _
"                {WM~SET_STATE, WM~MOUSE, 2}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_msg"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 1[mm],1[mm],1[mm],-11[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}," & vbLf & _
"                {WM~SET_FEATURE, WM~NO_DRAW, TRUE}," & vbLf & _
"                {WM~SET_FEATURE, WM~SEQUENCE, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~FRAME, ""fr_status"", ""fr_main""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],0[mm],0[mm],-13[mm]}," & vbLf & _
"                {WM~SET_FEATURE, WM~BORDER, ""BLACK_LINE""}," & vbLf & _
"                {WM~SET_FEATURE, WM~NO_DRAW, TRUE}" & vbLf & _
"            }," & vbLf & _
"            {WM~SLIDER, ""sld_status"", ""fr_status""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 1[mm],1[mm],1[mm],-7[mm]}," & vbLf & _
"                {WM~SET_VALUE, {0,100,1,0}}," & vbLf & _
"                {WM~SET_FEATURE, WM~BORDER, ""GLOW""}," & vbLf & _
"                {WM~SET_FEATURE, WM~INTERACTIVE, FALSE}" & vbLf & _
"            }," & vbLf & _
"            {WM~LABEL, ""lbl_status"", ""fr_status""," & vbLf & _
"                {WM~LEFT_BOTTOM_RIGHT_TOP, 1[mm],7[mm],1[mm],-12[mm]}," & vbLf & _
"                {WM~SET_VALUE, ""This is indicator message""}," & vbLf & _
"                {WM~SET_FEATURE, WM~CENTER, FALSE}" & vbLf & _
"            }} & get_tab_rotation( {""se_spec"",""ib_user"",""ib_pass"",""ib_start_date"",""ib_stop_date"",""pb_action""} ));" & vbLf & _
"" & vbLf & _
"        next_action := ""connect"";" & vbLf & _
"        is_jpn := get_locale = ""jpn-JPN"";" & vbLf & _
"        update_specs;" & vbLf & _
"        register_node( inst_availability, def );" & vbLf & _
"        send( !!""$$master_console"", {""QUIT_HANDSHAKE"",def,TRUE}, EVENT );" & vbLf & _
"    end INITIALIZE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define GET_DESTINATION_PATHNAME return PATH_OBJECT is" & vbLf & _
"        var" & vbLf & _
"            path : STRING;" & vbLf & _
"" & vbLf & _
"        loop" & vbLf & _
"            path := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-log_file-"" & unique & "".txt"";" & vbLf & _
"            next when fs_file_exists( path );" & vbLf & _
"            exit;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        return new_path( path );" & vbLf & _
"    end GET_DESTINATION_PATHNAME;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function GET_LINE_COUNT( text_file : STRING ) return NUMERIC is" & vbLf & _
"        var" & vbLf & _
"            cmd1       : STRING," & vbLf & _
"            cmd2       : STRING," & vbLf & _
"            tmp_file   : STRING," & vbLf & _
"            base_name  : STRING," & vbLf & _
"            tmp_script : STRING," & vbLf & _
"            line_count : WILDCARD;" & vbLf & _
"" & vbLf & _
"        -- define command" & vbLf & _
"        base_name  := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-get_line_count-"" & unique;" & vbLf & _
"        tmp_file   := base_name & "".txt"";" & vbLf & _
"        tmp_script := base_name & is_windows ? "".bat"" : "".sh"";" & vbLf & _
"        cmd1 := is_windows ? (""TYPE """""" & text_file & """""" | C:\Windows\System32\find.exe /C /V """""""" > """""" & tmp_file & """""" 2>&1"") :" & vbLf & _
"                             (""wc -l '"" & text_file & ""'  | awk '{print $1}' > '"" & tmp_file & ""' 2>&1"");" & vbLf & _
"        cmd2 := is_windows ? (""CMD /C """"""           & tmp_script &                  """""" & DEL """""" & tmp_script & """""""") :" & vbLf & _
"                             (""sh -c """"chmod a+x '"" & tmp_script & ""';'"" & tmp_script & ""';rm '""  & tmp_script & ""' "");" & vbLf & _
"" & vbLf & _
"        -- execute command" & vbLf & _
"        write( tmp_script, {cmd1} );" & vbLf & _
"        cf {FM~SPAWN_WAIT,  cmd2};" & vbLf & _
"" & vbLf & _
"        -- get line count" & vbLf & _
"        line_count := read( tmp_file );" & vbLf & _
"        line_count := s_eval( line_count(1) );" & vbLf & _
"        call( ""code_mgr"", {CM~DELETE_SOURCE, tmp_file}, EVENT );" & vbLf & _
"" & vbLf & _
"        return line_count;" & vbLf & _
"    end GET_LINE_COUNT;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function GET_LINES( text_file : STRING, start : NUMERIC, stop : NUMERIC ) return SET is" & vbLf & _
"        var" & vbLf & _
"            cmd1       : SET," & vbLf & _
"            lines      : SET," & vbLf & _
"            cmd2       : STRING," & vbLf & _
"            tmp_file   : STRING," & vbLf & _
"            base_name  : STRING," & vbLf & _
"            tmp_script : STRING;" & vbLf & _
"" & vbLf & _
"        -- define command" & vbLf & _
"        base_name  := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-get_lines-"" & unique;" & vbLf & _
"        tmp_file   := base_name & "".txt"";" & vbLf & _
"        tmp_script := base_name & is_windows ? "".bat"" : "".sh"";" & vbLf & _
"        cmd1 := is_windows ? {" & vbLf & _
"                                 ""@ECHO OFF""," & vbLf & _
"                                 ""SETLOCAL ENABLEDELAYEDEXPANSION""," & vbLf & _
"                                 ""SET FILE=""       & text_file," & vbLf & _
"                                 ""SET LINE_START="" & [STRING](start)," & vbLf & _
"                                 ""SET LINE_STOP=""  & [STRING](stop )," & vbLf & _
"                                 ""SET /A LINES = 0""," & vbLf & _
"                                 ""SET /A SKIP  = LINE_START - 1""," & vbLf & _
"                                 ""SET /A MAX   = LINE_STOP  - LINE_START + 1""," & vbLf & _
"                                 ""SET OPTIONS=""""usebackq delims=""""""," & vbLf & _
"                                 ""IF """"%SKIP%"""" GTR """"0"""" SET OPTIONS=""""usebackq skip=%SKIP% delims=""""""," & vbLf & _
"                                 ""FOR /F """"tokens=2 delims=:"""" %%a IN ('CHCP') DO @SET /A """"CP=%%~a""""""," & vbLf & _
"                                 ""CHCP 65001 > NUL""," & vbLf & _
"                                 ""FOR /F %OPTIONS% %%i IN (""""%FILE%"""") DO (""," & vbLf & _
"                                 ""    SET LINE=%%i""," & vbLf & _
"                                 ""    ECHO !LINE:--=^"" & LBL_DOUBLE_MINUS & ""!""," & vbLf & _
"                                 ""    SET /A LINES += 1""," & vbLf & _
"                                 ""    IF """"!LINES!"""" EQU """"%MAX%"""" GOTO EXIT""," & vbLf & _
"                                 "")""," & vbLf & _
"                                 "":EXIT""," & vbLf & _
"                                 ""    CHCP %CP% > NUL""," & vbLf & _
"                                 ""    ENDLOCAL""" & vbLf & _
"                             } : {" & vbLf & _
"                                 ""head -n "" & [STRING](stop) & "" '"" & text_file &" & vbLf & _
"                                 ""' | tail -n "" & [STRING](stop - start + 1) &" & vbLf & _
"                                 "" | sed """"s/--/\"" & LBL_DOUBLE_MINUS & ""/g"""""" &" & vbLf & _
"                                 "" > '"" & tmp_file & ""' 2>&1""" & vbLf & _
"                             };" & vbLf & _
"        cmd2 := is_windows ? (""CMD /C """"""           & tmp_script & """""" > """""" & tmp_file   & """""" & DEL """""" & tmp_script &  """""""") :" & vbLf & _
"                             (""sh -c """"chmod a+x '"" & tmp_script & ""';'""     & tmp_script &    ""';rm '""   & tmp_script & ""'"""""");" & vbLf & _
"" & vbLf & _
"        -- execute command" & vbLf & _
"        write( tmp_script, cmd1 );" & vbLf & _
"        cf {FM~SPAWN_WAIT, cmd2};" & vbLf & _
"" & vbLf & _
"        -- get lines" & vbLf & _
"        lines := read( tmp_file );" & vbLf & _
"        call( ""code_mgr"", {CM~DELETE_SOURCE, tmp_file}, EVENT );" & vbLf & _
"" & vbLf & _
"        return lines;" & vbLf & _
"    end GET_LINES;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function GET_DATE_RANGE( log_file : STRING, start : UNIVERSAL, stop : UNIVERSAL ) is" & vbLf & _
"        var" & vbLf & _
"            DATE_TIME_LENGTH : constant NUMERIC := 20," & vbLf & _
"" & vbLf & _
"            lines : SET," & vbLf & _
"            line  : STRING," & vbLf & _
"            count : NUMERIC," & vbLf & _
"            time  : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        -- get line count" & vbLf & _
"        return when log_file = NULL;" & vbLf & _
"        count := get_line_count( log_file );" & vbLf & _
"" & vbLf & _
"        -- start date time" & vbLf & _
"        line := (get_lines( log_file, 1, 1 ))(1);" & vbLf & _
"        if size( line ) > DATE_TIME_LENGTH then" & vbLf & _
"            time := date( nth( line, 1, DATE_TIME_LENGTH ) );" & vbLf & _
"            start := (start = 0[s] or else start > time) ? time : start;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- stop date time" & vbLf & _
"        line := (get_lines( log_file, count, count ))(1);" & vbLf & _
"        if size( line ) > DATE_TIME_LENGTH then" & vbLf & _
"            time := date( nth( line, 1, DATE_TIME_LENGTH ) );" & vbLf & _
"            stop := (stop = 0[s] or else stop < time) ? time : stop;" & vbLf & _
"        end if;" & vbLf & _
"    end GET_DATE_RANGE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function SAVE_CSV_FILE( csv_file : STRING, text : SET ) return WILDCARD is" & vbLf & _
"        var" & vbLf & _
"            script     : SET," & vbLf & _
"            id         : STRING," & vbLf & _
"            cmd        : STRING," & vbLf & _
"            tmp_file   : STRING," & vbLf & _
"            txt_file   : STRING," & vbLf & _
"            tmp_script : STRING," & vbLf & _
"            result     : WILDCARD;" & vbLf & _
"" & vbLf & _
"        -- save csv file" & vbLf & _
"        id := unique;" & vbLf & _
"        tmp_file   := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-save_csv_file-"" & id & "".csv"";" & vbLf & _
"        txt_file   := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-save_csv_file-"" & id & "".txt"";" & vbLf & _
"        tmp_script := string_pref( ""TMP"" ) & ""delta-machine_time_calculator-save_csv_file-"" & id & (is_windows ? "".vbs"" : "".sh"");" & vbLf & _
"        result     := write( tmp_file, !text );" & vbLf & _
"        return result when type result = [STATUS];" & vbLf & _
"" & vbLf & _
"        -- write script for fixing csv file character encoding" & vbLf & _
"        script := is_windows ?" & vbLf & _
"            {""' script for fixing UTF-8 encoded text file""," & vbLf & _
"             """"," & vbLf & _
"             ""' constants ------------------------------------------------""," & vbLf & _
"             ""Const ForReading   = 1""," & vbLf & _
"             ""Const ForWriting   = 2""," & vbLf & _
"             ""Const adTypeBinary = 1""," & vbLf & _
"             ""Const adTypeText   = 2""," & vbLf & _
"             ""Const adSaveCreateOverWrite = 2""," & vbLf & _
"             """"," & vbLf & _
"             ""' global variables -----------------------------------------""," & vbLf & _
"             ""Set objFileSystem = CreateObject(""""Scripting.FileSystemObject"""")""," & vbLf & _
"             ""TXT_FILE = """""" & txt_file & """"""""," & vbLf & _
"             ""SRC_FILE = """""" & tmp_file & """"""""," & vbLf & _
"             ""DST_FILE = """""" & csv_file & """"""""," & vbLf & _
"             """"," & vbLf & _
"             ""' functions ------------------------------------------------""," & vbLf & _
"             ""Function HasNonAsciiCharacters(strFilename)""," & vbLf & _
"             ""  ' check non-ascii character exists in strFilename""," & vbLf & _
"             ""  HasNonAsciiCharacters = False""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' define pattern for the detection""," & vbLf & _
"             ""  Set objRegExp = new RegExp""," & vbLf & _
"             ""  objRegExp.Pattern = """"[^\x01-\x7F]""""""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' read text file""," & vbLf & _
"             ""  Set objFile = objFileSystem.OpenTextFile(strFilename, ForReading)""," & vbLf & _
"             ""  Do Until objFile.AtEndOfStream""," & vbLf & _
"             ""    strLine = objFile.ReadLine""," & vbLf & _
"             ""    If objRegExp.Test(strLine) Then""," & vbLf & _
"             ""      ' non-ascii character is detected""," & vbLf & _
"             ""      HasNonAsciiCharacters = True""," & vbLf & _
"             ""      Exit Do""," & vbLf & _
"             ""    End If""," & vbLf & _
"             ""  Loop""," & vbLf & _
"             ""  objFile.Close""," & vbLf & _
"             ""End Function""," & vbLf & _
"             """"," & vbLf & _
"             ""Sub ConvertTextFile(strSrcFilename, strDstFilename)""," & vbLf & _
"             ""  ' convert UTF-8 encoded text file to add BOM at the top of the file""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' load UTF-8 text file without BOM""," & vbLf & _
"             ""  Set objStream = CreateObject(""""ADODB.Stream"""")""," & vbLf & _
"             ""  objStream.Open""," & vbLf & _
"             ""  objStream.Type = adTypeText""," & vbLf & _
"             ""  objStream.Charset = """"UTF-8""""""," & vbLf & _
"             ""  objStream.LoadFromFile strSrcFilename""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' save UTF-8 text file with BOM""," & vbLf & _
"             ""  objStream.SaveTofile strDstFilename, adSaveCreateOverWrite""," & vbLf & _
"             ""  objStream.Close""," & vbLf & _
"             ""End Sub""," & vbLf & _
"             """"," & vbLf & _
"             ""Sub WriteTextFile(strMessage, strFilename)""," & vbLf & _
"             ""  ' write UTF-8 text file without BOM""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' convert into UTF-8 text""," & vbLf & _
"             ""  Set objStream = CreateObject(""""ADODB.Stream"""")""," & vbLf & _
"             ""  objStream.Open""," & vbLf & _
"             ""  objStream.Type = adTypeText""," & vbLf & _
"             ""  objStream.Charset = """"UTF-8""""""," & vbLf & _
"             ""  objStream.WriteText strMessage""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' read UTF-8 text without BOM""," & vbLf & _
"             ""  objStream.Position = 0""," & vbLf & _
"             ""  objStream.Type = adTypeBinary""," & vbLf & _
"             ""  objStream.Position = 3""," & vbLf & _
"             ""  binText = objStream.Read""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' save UTF-8 text into strFilename""," & vbLf & _
"             ""  objStream.Position = 0""," & vbLf & _
"             ""  objStream.Write binText""," & vbLf & _
"             ""  objStream.SetEOS""," & vbLf & _
"             ""  objStream.SaveToFile strFilename, adSaveCreateOverWrite""," & vbLf & _
"             ""  objStream.Close""," & vbLf & _
"             ""End Sub""," & vbLf & _
"             """"," & vbLf & _
"             ""Sub FixTextFile""," & vbLf & _
"             ""  ' fix UTF-8 encoded text file""," & vbLf & _
"             ""  On Error Resume Next""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' check non-ascii character exists in SRC_FILE""," & vbLf & _
"             ""  If HasNonAsciiCharacters(SRC_FILE) Then""," & vbLf & _
"             ""    ' SRC_FILE has non-ascii character""," & vbLf & _
"             ""    ' add BOM at the top of the text file""," & vbLf & _
"             ""    ConvertTextFile SRC_FILE, DST_FILE""," & vbLf & _
"             ""  Else""," & vbLf & _
"             ""    ' SRC_FILE does not have non-ascii character""," & vbLf & _
"             ""    ' just copy SRC_FILE to DST_FILE""," & vbLf & _
"             ""    objFileSystem.CopyFile SRC_FILE, DST_FILE, True""," & vbLf & _
"             ""  End If""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  ' check error""," & vbLf & _
"             ""  If Err.Number <> 0 Then""," & vbLf & _
"             ""    strMessage = """"Error Code : """" & Err.Number & """", Description : """" & Err.Description""," & vbLf & _
"             ""    WriteTextFile strMessage, TXT_FILE""," & vbLf & _
"             ""  End If""," & vbLf & _
"             ""End Sub""," & vbLf & _
"             """"," & vbLf & _
"             ""' main procedure -------------------------------------------""," & vbLf & _
"             ""FixTextFile""} :" & vbLf & _
"            {""# script for fixing UTF-8 encoded text file""," & vbLf & _
"             """"," & vbLf & _
"             ""# global variables -----------------------------------------""," & vbLf & _
"             ""SRC_FILE="""""" & tmp_file & """"""""," & vbLf & _
"             ""DST_FILE="""""" & csv_file & """"""""," & vbLf & _
"             """"," & vbLf & _
"             ""# functions ------------------------------------------------""," & vbLf & _
"             ""function fix_text_file""," & vbLf & _
"             ""{""," & vbLf & _
"             ""  # fix UTF-8 encoded text file""," & vbLf & _
"             ""  ""," & vbLf & _
"             ""  # check non-ascii character exists in SRC_FILE""," & vbLf & _
"             ""  line=`grep $'[^\x1-\x7F]' """"$SRC_FILE""""`""," & vbLf & _
"             ""  if [ """"$line"""" != """""""" ]; then""," & vbLf & _
"             ""    # SRC_FILE has non-ascii character""," & vbLf & _
"             ""    # add BOM at the top of the text file""," & vbLf & _
"             ""    printf '\xEF\xBB\xBF' > """"$DST_FILE""""""," & vbLf & _
"             ""    cat    """"$SRC_FILE""""   >> """"$DST_FILE""""""," & vbLf & _
"             ""  else""," & vbLf & _
"             ""    # SRC_FILE does not have non-ascii character""," & vbLf & _
"             ""    # just copy SRC_FILE to DST_FILE""," & vbLf & _
"             ""    cp """"$SRC_FILE"""" """"$DST_FILE""""""," & vbLf & _
"             ""  fi""," & vbLf & _
"             ""}""," & vbLf & _
"             """"," & vbLf & _
"             ""# main procedure -------------------------------------------""," & vbLf & _
"             ""fix_text_file""};" & vbLf & _
"        write( tmp_script, script );" & vbLf & _
"" & vbLf & _
"        -- fix csv file character encoding" & vbLf & _
"        cmd := (is_windows ? ""CMD /C"" : ""sh"") & "" """""" & tmp_script & """""""";" & vbLf & _
"        cf {FM~SPAWN_WAIT, cmd};" & vbLf & _
"        call( ""code_mgr"", {CM~DELETE_SOURCE, tmp_file  }, EVENT );" & vbLf & _
"        call( ""code_mgr"", {CM~DELETE_SOURCE, tmp_script}, EVENT );" & vbLf & _
"" & vbLf & _
"        -- check error message text file" & vbLf & _
"        if cf {FM~FILE_EXISTS, txt_file} then" & vbLf & _
"            text := read( txt_file );" & vbLf & _
"            call( ""code_mgr"", {CM~DELETE_SOURCE, txt_file}, EVENT );" & vbLf & _
"            return text;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- check csv_file" & vbLf & _
"        return SC~F~UNHANDLED_FAILURE when (cf {FM~FILE_EXISTS, csv_file}) = FALSE;" & vbLf & _
"        return csv_file;" & vbLf & _
"    end SAVE_CSV_FILE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_DATE_STRING( year : NUMERIC, month : NUMERIC, day : NUMERIC ) return STRING is" & vbLf & _
"        if is_jpn then" & vbLf & _
"            -- return date string ""2017/12/26""" & vbLf & _
"            return format( year,  ""z9999"" ) & ""/"" &" & vbLf & _
"                   format( month, ""z99""   ) & ""/"" &" & vbLf & _
"                   format( day,   ""z99""   );" & vbLf & _
"        else" & vbLf & _
"            -- return date string ""26-DEC-2017""" & vbLf & _
"            return format( day,   ""z99""   ) & ""-"" &" & vbLf & _
"                   month_abbrev(month)      & ""-"" &" & vbLf & _
"                   format( year,  ""z9999"" );" & vbLf & _
"        end if;" & vbLf & _
"    end TO_DATE_STRING;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_DATE_STRING( time : UNIVERSAL ) return STRING is" & vbLf & _
"        var" & vbLf & _
"            str : STRING := date( time );" & vbLf & _
"" & vbLf & _
"        if is_jpn then" & vbLf & _
"            -- convert ""26-DEC-2017 11:22:06"" into ""2017/12/26""" & vbLf & _
"            return to_date_string( s_eval( nth( str, 8, 4 ) )," & vbLf & _
"                                    index( nth( str, 4, 3 ), month_abbrev )," & vbLf & _
"                                   s_eval( nth( str, 1, 2 ) ) );" & vbLf & _
"        else" & vbLf & _
"            -- convert ""26-DEC-2017 11:22:06"" into ""26-DEC-2017""" & vbLf & _
"            str := nth( str, 1, DATE_LENGTH_ENG);" & vbLf & _
"            str := (str(1) = "" "") ? (""0"" & remove( str, 1 )) : str;" & vbLf & _
"            return str;" & vbLf & _
"        end if;" & vbLf & _
"    end TO_DATE_STRING;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_DATE_TIME( str_date : STRING, str_time : STRING := ""00:00:00"" ) return UNIVERSAL is" & vbLf & _
"        if is_jpn then" & vbLf & _
"            -- convert ""2017/12/26"" into ""26-DEC-2017 00:00:00""" & vbLf & _
"            return NULL when size( str_date ) /= DATE_LENGTH_JPN;" & vbLf & _
"            str_date := nth( str_date, 9, 2 ) & ""-"" & month_abbrev(s_eval( nth( str_date, 6, 2 ) )) & ""-"" & nth( str_date, 1, 4 );" & vbLf & _
"            return date( str_date & "" "" & str_time );" & vbLf & _
"        else" & vbLf & _
"            -- convert ""26-DEC-2017"" into ""26-DEC-2017 00:00:00""" & vbLf & _
"            return NULL when size( str_date ) /= DATE_LENGTH_ENG;" & vbLf & _
"            return date( str_date & "" "" & str_time );" & vbLf & _
"        end if;" & vbLf & _
"    end TO_DATE_TIME;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_DATE_TIME_STRING( time : UNIVERSAL ) return STRING is" & vbLf & _
"        var" & vbLf & _
"            str : STRING := date( time );" & vbLf & _
"" & vbLf & _
"        -- correct "" 1-DEC-2017 00:00:00"" -> ""1-DEC-2017 00:00:00""" & vbLf & _
"        -- so that EXCEL can identify the string is a date time." & vbLf & _
"        return (str(1) = "" "") ? remove( str, 1, 1 ) : str;" & vbLf & _
"    end TO_DATE_TIME_STRING;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_DURATION_STRING( duration : UNIVERSAL ) return STRING is" & vbLf & _
"        var" & vbLf & _
"            SECOND : constant UNIVERSAL := 1[s]," & vbLf & _
"            MINUTE : constant UNIVERSAL := 60 * SECOND," & vbLf & _
"            HOUR   : constant UNIVERSAL := 60 * MINUTE," & vbLf & _
"            DAY    : constant UNIVERSAL := 24 * HOUR," & vbLf & _
"" & vbLf & _
"            time : NUMERIC," & vbLf & _
"            str  : STRING;" & vbLf & _
"" & vbLf & _
"        -- convert 131696[s] into ""1 days 12:34:56""" & vbLf & _
"        for length in {DAY,HOUR,MINUTE,SECOND} loop" & vbLf & _
"            time := lower( duration / length );" & vbLf & _
"            duration := duration - length * time;" & vbLf & _
"            if length = DAY then" & vbLf & _
"                str := [STRING](time) & "" days"";" & vbLf & _
"            else" & vbLf & _
"                str := str & ((length = HOUR) ? "" "" : "":"") & format( time, ""z99"" );" & vbLf & _
"            end if;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        return str;" & vbLf & _
"    end TO_DURATION_STRING;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function TO_STRING( val : WILDCARD, default : STRING := """" ) return STRING is" & vbLf & _
"        return default when val = NULL;" & vbLf & _
"        return [STRING]( val );" & vbLf & _
"    end TO_STRING;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function DOWNLOAD_FILE( src : PATH_OBJECT, dst : PATH_OBJECT ) return STATUS is" & vbLf & _
"        var" & vbLf & _
"            WAIT_INTERVAL : constant UNIVERSAL := 0.5[s]," & vbLf & _
"" & vbLf & _
"            result      : STATUS," & vbLf & _
"            msg         : WILDCARD," & vbLf & _
"            node_child  : STRING := ""MACHINE_TIME_CALCULATOR:DOWNLOAD_FILE:CHILD:""  & unique," & vbLf & _
"            node_parent : STRING := ""MACHINE_TIME_CALCULATOR:DOWNLOAD_FILE:PARENT:"" & unique;" & vbLf & _
"" & vbLf & _
"        -- download a log file. this may take minutes ..." & vbLf & _
"        spawn block" & vbLf & _
"            result := fs_download_file( src, dst );" & vbLf & _
"            msg    := receive( node_child, EVENT, NULL, NULL, FALSE );" & vbLf & _
"            if msg = SC~F~NO_DATA_RETURNED then" & vbLf & _
"                -- fs_download_file finished. notify parent node." & vbLf & _
"                send( node_parent, {""finished"", result}, EVENT );" & vbLf & _
"            elsif result = SC~S~SUCCESS then" & vbLf & _
"                -- fs_download_file was aborted. delete downloaded file" & vbLf & _
"                call( ""code_mgr"", {CM~DELETE_SOURCE, dst::get_path()}, EVENT );" & vbLf & _
"            end if;" & vbLf & _
"            return;" & vbLf & _
"        end block;" & vbLf & _
"" & vbLf & _
"        -- wait for fs_download_file" & vbLf & _
"        loop" & vbLf & _
"            msg := receive( node_parent, EVENT, NULL, NULL, FALSE );" & vbLf & _
"            if type msg = [SET] and then msg(1) = ""finished"" then" & vbLf & _
"                -- fs_download_file finished" & vbLf & _
"                result := msg(2);" & vbLf & _
"                exit;" & vbLf & _
"            elsif handle_state( ""is aborted"" ) then" & vbLf & _
"                -- abort current download" & vbLf & _
"                send( node_child, {""aborted""}, EVENT );" & vbLf & _
"                result := SC~F~OPERATION_ABORTED;" & vbLf & _
"                exit;" & vbLf & _
"            end if;" & vbLf & _
"            delay WAIT_INTERVAL;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        return result;" & vbLf & _
"    end DOWNLOAD_FILE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define CONNECT is" & vbLf & _
"        -- connect to the spectrometer and download log files" & vbLf & _
"        var" & vbLf & _
"            text   : SET," & vbLf & _
"            files  : SET," & vbLf & _
"            fname  : STRING," & vbLf & _
"            idx    : NUMERIC," & vbLf & _
"            rate   : NUMERIC," & vbLf & _
"            count  : NUMERIC," & vbLf & _
"            steps  : NUMERIC," & vbLf & _
"            result : WILDCARD," & vbLf & _
"            start  : UNIVERSAL," & vbLf & _
"            stop   : UNIVERSAL," & vbLf & _
"            dst    : PATH_OBJECT," & vbLf & _
"            src    : PATH_OBJECT," & vbLf & _
"            path   : PATH_OBJECT," & vbLf & _
"            specs  : SET     := find_devices( ""JEOL Spectrometer"" )," & vbLf & _
"            spec   : STRING  := cw {{def,""se_spec""}, {WM~GET_SELECTED}}," & vbLf & _
"            user   : STRING  := cw {{def,""ib_user""}, {WM~GET_SELECTED}}," & vbLf & _
"            pass   : STRING  := cw {{def,""ib_pass""}, {WM~GET_SELECTED}};" & vbLf & _
"" & vbLf & _
"        -- connect to the spectrometer" & vbLf & _
"        handle_state( ""enable abort"" );" & vbLf & _
"        update_indicator( ""update"", 10, ""Now connecting to the spectrometer ..."" );" & vbLf & _
"        if nc /= null_netcontext then" & vbLf & _
"            fs_disconnect( nc );" & vbLf & _
"            nc := null_netcontext;" & vbLf & _
"        end if;" & vbLf & _
"        for dev in specs loop" & vbLf & _
"            if dev.DEVICE_DESC.site = spec then" & vbLf & _
"                nc := dev.DEVICE_DESC.nc;" & vbLf & _
"                exit;" & vbLf & _
"            end if;" & vbLf & _
"        end loop;" & vbLf & _
"        nc := set_password( set_user( nc, user ), pass );" & vbLf & _
"        result := fs_connect( nc );" & vbLf & _
"        if result /= SC~S~SUCCESS then" & vbLf & _
"            show_message( (result = SC~F~NO_PRIVILEGE) ? {""Failed to connect to the spectrometer.""," & vbLf & _
"                                                          ""Please check username and password.""} :" & vbLf & _
"                                                         {""Failed to connect to the spectrometer : "" & [STRING](result)," & vbLf & _
"                                                          ""Please check network environment.""}," & vbLf & _
"                          MSG_ERROR, ""enable connect"" );" & vbLf & _
"            return;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- check the privilege" & vbLf & _
"        update_indicator( ""update"", 20, ""Now checking the privilege ..."" );" & vbLf & _
"        if not verify_privilege( nc, ""CONSOLE"" ) then" & vbLf & _
"            show_message( {""User '"" & user & ""' does not have 'CONSOLE' privilege.""," & vbLf & _
"                           ""Please type username and password who has 'CONSOLE' privilege.""}," & vbLf & _
"                          MSG_ERROR, ""enable connect"" );" & vbLf & _
"            return;" & vbLf & _
"        elsif not verify_privilege( nc, ""READ"" ) then" & vbLf & _
"            show_message( {""User '"" & user & ""' does not have 'READ' privilege.""," & vbLf & _
"                           ""Please type username and password who has 'READ' privilege.""}," & vbLf & _
"                          MSG_ERROR, ""enable connect"" );" & vbLf & _
"            return;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- check folder_table" & vbLf & _
"        update_indicator( ""update"", 25, ""Now checking folder_table ..."" );" & vbLf & _
"        result := query( nc, ""SELECT tbl_name FROM sqlite_master WHERE type='table'"");" & vbLf & _
"        if type result = [STATUS] or else" & vbLf & _
"          (type result = [SET] and then size( result ) > 0 and then type result(1) = [STATUS]) then" & vbLf & _
"            show_message( {""Failed to access Data Server : "" & [STRING](result)," & vbLf & _
"                           ""Please restart control service and try again.""}," & vbLf & _
"                          MSG_ERROR, ""enable connect"" );" & vbLf & _
"            return;" & vbLf & _
"        else" & vbLf & _
"            use_folder := (type result = [SET]) ? ({{""folder_table""}} in result) : FALSE;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- check log files" & vbLf & _
"        update_indicator( ""update"", 30, ""Now checking log files ..."" );" & vbLf & _
"        logs  := repeat( {{}}, size( BASENAMES ) );" & vbLf & _
"        path  := new_spectrometer_path( nc, ""logs/"" );" & vbLf & _
"        files := path::get_file_list();" & vbLf & _
"        count := 0;" & vbLf & _
"        for i : NUMERIC is 1 to size( files ) loop" & vbLf & _
"            fname := replace_string( get_basename( files(i) ), ""-old"", """" );" & vbLf & _
"            idx   := index( fname, BASENAMES );" & vbLf & _
"            next when idx = 0;" & vbLf & _
"            logs(idx) := logs(idx) & {files(i)};" & vbLf & _
"            count := count + 1;" & vbLf & _
"        end loop;" & vbLf & _
"        for i : NUMERIC is 1 to size( BASENAMES ) loop" & vbLf & _
"            next when size( logs(i) ) > 0;" & vbLf & _
"            show_message( {""'"" & BASENAMES(i) & "".log' does not exist.""," & vbLf & _
"                           ""Machine time could not be calculated, sorry.""}," & vbLf & _
"                          MSG_ERROR, ""enable connect"" );" & vbLf & _
"            return;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        -- download log files" & vbLf & _
"        rate  := 30;" & vbLf & _
"        steps := (90 - rate) / count;" & vbLf & _
"        for i : NUMERIC is 1 to size( logs ) loop" & vbLf & _
"            for j : NUMERIC is 1 to size( logs(i) ) loop" & vbLf & _
"                rate  := rate + steps;" & vbLf & _
"                src   := logs(i,j);" & vbLf & _
"                dst   := get_destination_pathname;" & vbLf & _
"                fname := src::get_name();" & vbLf & _
"                update_indicator( ""update"", rate, ""Now downloading '"" & fname & ""' ..."" );" & vbLf & _
"                result := download_file( src, dst );" & vbLf & _
"                if handle_state( ""is aborted"" ) then" & vbLf & _
"                    show_message( ""Downloading log files were cancelled."", MSG_WARNING, ""enable connect"" );" & vbLf & _
"                    return;" & vbLf & _
"                elsif result /= SC~S~SUCCESS then" & vbLf & _
"                    show_message( {""Failed to download '"" & fname & ""' : "" & [STRING](result)," & vbLf & _
"                                   ""Machine time could not be calculated, sorry.""}," & vbLf & _
"                                  MSG_ERROR, ""enable connect"" );" & vbLf & _
"                    return;" & vbLf & _
"                end if;" & vbLf & _
"                logs(fname) := dst::get_path();" & vbLf & _
"                temp_files  := temp_files & {dst::get_path()};" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        -- check machine.log date range" & vbLf & _
"        start := 0[s];" & vbLf & _
"        stop  := 0[s];" & vbLf & _
"        update_indicator( ""update"", 100, ""Now checking date range ..."" );" & vbLf & _
"        get_date_range( logs(""machine-old.log""), @start, @stop );" & vbLf & _
"        get_date_range( logs(""machine.log""),     @start, @stop );" & vbLf & _
"        start_date   := to_date_string( start );" & vbLf & _
"        stop_date    := to_date_string( stop  );" & vbLf & _
"        target_field := cw {{def,""se_target""}, {WM~GET_SELECTED}};" & vbLf & _
"" & vbLf & _
"        -- enable calculation" & vbLf & _
"         log_message( ""Downloaded "" & [STRING](count) & "" log files successfully."", MSG_INFO );" & vbLf & _
"        show_message( ""Please check [Start Date]/[Stop Date] and click [Calculate] button."", MSG_INFO, ""enable calculate"" );" & vbLf & _
"        next_action := ""calculate"";" & vbLf & _
"    end CONNECT;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define CALCULATE is" & vbLf & _
"        var" & vbLf & _
"            text       : SET," & vbLf & _
"            text_file  : STRING," & vbLf & _
"            line_idx   : NUMERIC," & vbLf & _
"            line_max   : NUMERIC," & vbLf & _
"            line_stop  : NUMERIC," & vbLf & _
"            line_start : NUMERIC," & vbLf & _
"            line_count : NUMERIC;" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        function INIT_LINE_INFO( file_path : STRING ) is" & vbLf & _
"            text_file  := file_path;" & vbLf & _
"            line_idx   := 0;" & vbLf & _
"            line_max   := 0;" & vbLf & _
"            line_stop  := 0;" & vbLf & _
"            line_count := get_line_count( text_file );" & vbLf & _
"        end INIT_LINE_INFO;" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        define GET_NEXT_LINE return STRING is" & vbLf & _
"            var" & vbLf & _
"                MAX_TEXT_LINES : constant NUMERIC := 6000;" & vbLf & _
"" & vbLf & _
"            if line_idx + 1 > line_max then" & vbLf & _
"                line_start := line_stop + 1;" & vbLf & _
"                line_stop  := min( line_count, line_start + MAX_TEXT_LINES );" & vbLf & _
"                return """" when line_stop < line_start;" & vbLf & _
"" & vbLf & _
"                text := get_lines( text_file, line_start, line_stop );" & vbLf & _
"                line_max := size( text );" & vbLf & _
"                line_idx := 0;" & vbLf & _
"                return """" when line_idx + 1 > line_max;" & vbLf & _
"            end if;" & vbLf & _
"" & vbLf & _
"            line_idx := line_idx + 1;" & vbLf & _
"            return replace_string( text( line_idx ), LBL_DOUBLE_MINUS, ""--"" );" & vbLf & _
"        end GET_NEXT_LINE;" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        define PREVIEW_NEXT_LINE return STRING is" & vbLf & _
"            return (line_idx + 1 > line_max) ? """" : replace_string( text( line_idx + 1 ), LBL_DOUBLE_MINUS, ""--"" );" & vbLf & _
"        end PREVIEW_NEXT_LINE;" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        var" & vbLf & _
"            MAX_REC_SIZE       : constant NUMERIC   := 50," & vbLf & _
"            POS_LINE_START     : constant NUMERIC   :=  1," & vbLf & _
"            POS_TIME_END       : constant NUMERIC   := 20," & vbLf & _
"            POS_JOB_START      : constant NUMERIC   := 43," & vbLf & _
"            POS_FATAL_END      : constant NUMERIC   := 49," & vbLf & _
"            POS_USER_START     : constant NUMERIC   := 60," & vbLf & _
"            UPDATE_INTERVAL    : constant UNIVERSAL :=  1[s]," & vbLf & _
"            TIME_TOLERANCE     : constant UNIVERSAL :=  2[s]," & vbLf & _
"            JOB_TIME_TOLERANCE : constant UNIVERSAL := 45[s]," & vbLf & _
"" & vbLf & _
"            pos         : SET," & vbLf & _
"            row         : SET," & vbLf & _
"            files       : SET," & vbLf & _
"            steps       : SET," & vbLf & _
"            result      : SET," & vbLf & _
"            sec_op      : SET," & vbLf & _
"            sec_exp     : SET," & vbLf & _
"            sec_owner   : SET," & vbLf & _
"            data        : SET," & vbLf & _
"            data_op     : SET," & vbLf & _
"            data_exp    : SET," & vbLf & _
"            data_proj   : SET," & vbLf & _
"            data_user   : SET," & vbLf & _
"            data_owner  : SET," & vbLf & _
"            data_folder : SET," & vbLf & _
"            idxs_list   : SET," & vbLf & _
"            curr_list   : SET," & vbLf & _
"            proj_list   : SET," & vbLf & _
"            user_list   : SET," & vbLf & _
"            start_list  : SET," & vbLf & _
"            folder_list : SET," & vbLf & _
"            line        : STRING," & vbLf & _
"            state       : STRING," & vbLf & _
"            folder      : STRING," & vbLf & _
"            exp_file    : STRING," & vbLf & _
"            username    : STRING," & vbLf & _
"            job_name    : STRING," & vbLf & _
"            job_result  : STRING," & vbLf & _
"            proj_name   : STRING," & vbLf & _
"            name_id     : STRING," & vbLf & _
"            data_id     : STRING," & vbLf & _
"            query_text  : STRING," & vbLf & _
"            sz          : NUMERIC," & vbLf & _
"            rate        : NUMERIC," & vbLf & _
"            idx         : NUMERIC," & vbLf & _
"            idx1        : NUMERIC," & vbLf & _
"            idx2        : NUMERIC," & vbLf & _
"            idx_op      : NUMERIC," & vbLf & _
"            idx_exp     : NUMERIC," & vbLf & _
"            idx_proj    : NUMERIC," & vbLf & _
"            idx_user    : NUMERIC," & vbLf & _
"            idx_owner   : NUMERIC," & vbLf & _
"            idx_folder  : NUMERIC," & vbLf & _
"            prev        : UNIVERSAL," & vbLf & _
"            op_stop     : UNIVERSAL," & vbLf & _
"            exp_stop    : UNIVERSAL," & vbLf & _
"            exp_start   : UNIVERSAL," & vbLf & _
"            job_date    : UNIVERSAL," & vbLf & _
"            curr_time   : UNIVERSAL," & vbLf & _
"            stop_time   : UNIVERSAL," & vbLf & _
"            start_time  : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        -- initialize" & vbLf & _
"        conf            := {};" & vbLf & _
"        data_by_user    := {};" & vbLf & _
"        data_by_proj    := {};" & vbLf & _
"        data_by_folder  := {};" & vbLf & _
"        table_by_user   := {};" & vbLf & _
"        table_by_proj   := {};" & vbLf & _
"        table_by_folder := {};" & vbLf & _
"        owner_sections  := {};" & vbLf & _
"        folder_used     := FALSE;" & vbLf & _
"        handle_state( ""enable abort"", ""Calculating machine time was cancelled."", ""enable calculate"" );" & vbLf & _
"" & vbLf & _
"        -- parse machine.log file" & vbLf & _
"        rate  := 0;" & vbLf & _
"        prev  := seconds;" & vbLf & _
"        state := ""find ownership began"";" & vbLf & _
"        files := ((logs(""machine-old.log"") = NULL) ? {} : {""machine-old.log""}) & {""machine.log""};" & vbLf & _
"        steps := {100 / size( files ), NULL};" & vbLf & _
"        start_time := to_date_time( start_date, ""00:00:00"" );" & vbLf & _
"        stop_time  := to_date_time(  stop_date, ""23:59:59"" );" & vbLf & _
"        update_indicator( ""update"", 0, ""Now parsing machine.log file ..."" );" & vbLf & _
"" & vbLf & _
"        for fl in files loop" & vbLf & _
"            init_line_info( logs(fl) );" & vbLf & _
"            steps(2) := steps(1) / line_count;" & vbLf & _
"" & vbLf & _
"            for i : NUMERIC is 1 to line_count loop" & vbLf & _
"                rate := rate + steps(2);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"                line := get_next_line;" & vbLf & _
"                curr_time := date( nth( line, POS_LINE_START, POS_TIME_END ) );" & vbLf & _
"" & vbLf & _
"                -- reset current state for boot message" & vbLf & _
"                if ""Network Services Active"" in line then" & vbLf & _
"                    state := ""find ownership began"";" & vbLf & _
"                    next;" & vbLf & _
"                end if;" & vbLf & _
"" & vbLf & _
"                case state is" & vbLf & _
"                    when ""find ownership began"" =>" & vbLf & _
"                        if ""INFO : Instrument ownership began for"" in line then" & vbLf & _
"                            sec_owner := new OWNER_SECTION;" & vbLf & _
"                            sec_owner.OWNER_SECTION.start       := curr_time;" & vbLf & _
"                            sec_owner.OWNER_SECTION.username    := nth( line, POS_USER_START, index( "" from "", line ) - POS_USER_START );" & vbLf & _
"                            sec_owner.OWNER_SECTION.op_sections := {};" & vbLf & _
"" & vbLf & _
"                            exit when sec_owner.OWNER_SECTION.start > stop_time;" & vbLf & _
"                            if target_field = ""Operation Began"" or else sec_owner.OWNER_SECTION.start >= start_time then" & vbLf & _
"                                state := ""find operation began"";" & vbLf & _
"                            end if;" & vbLf & _
"                        elsif ""INFO : Previous instrument operation ended for"" in line and then" & vbLf & _
"                              curr_time >= start_time and then curr_time <= stop_time and then size( sec_owner ) = 0 then" & vbLf & _
"                            line := preview_next_line;" & vbLf & _
"                            if ""INFO : Instrument operation began for"" in line then" & vbLf & _
"                                sec_owner := new OWNER_SECTION;" & vbLf & _
"                                sec_owner.OWNER_SECTION.start       := curr_time;" & vbLf & _
"                                sec_owner.OWNER_SECTION.username    := nth( line, POS_USER_START, index( "" from "", line ) - POS_USER_START );" & vbLf & _
"                                sec_owner.OWNER_SECTION.op_sections := {};" & vbLf & _
"                                state := ""find operation began"";" & vbLf & _
"                            end if;" & vbLf & _
"                        end if;" & vbLf & _
"" & vbLf & _
"                    when ""find operation began"" =>" & vbLf & _
"                        if ""INFO : Instrument operation began for"" in line and then" & vbLf & _
"                           (target_field = ""Owner Began"" or else (curr_time >= start_time and then curr_time <= stop_time)) then" & vbLf & _
"                            sec_op := new OP_SECTION;" & vbLf & _
"                            sec_op.OP_SECTION.start        := curr_time;" & vbLf & _
"                            sec_op.OP_SECTION.exp_sections := {};" & vbLf & _
"                            state := ""find experiment start"";" & vbLf & _
"                        elsif ""INFO : Completed Job "" in line and then size( sec_owner.OWNER_SECTION.op_sections ) > 0 then" & vbLf & _
"                            idx := size( sec_owner.OWNER_SECTION.op_sections );" & vbLf & _
"                            sec_owner.OWNER_SECTION.op_sections(idx).OP_SECTION.job_result := ""FINISHED"";" & vbLf & _
"                        elsif ""FATAL : Job "" in line and then size( sec_owner.OWNER_SECTION.op_sections ) > 0 then" & vbLf & _
"                            idx := size( sec_owner.OWNER_SECTION.op_sections );" & vbLf & _
"                            sec_owner.OWNER_SECTION.op_sections(idx).OP_SECTION.job_result := ""FAILED"";" & vbLf & _
"                        elsif ""Terminating gradient shim.""        in line or else" & vbLf & _
"                              ""Gradient Shim: Completed in ""      in line or else" & vbLf & _
"                              ""CANCELLED : Aborted gradient shim"" in line then" & vbLf & _
"                            idx_op := size( sec_owner.OWNER_SECTION.op_sections );" & vbLf & _
"                            if idx_op > 0 and then" & vbLf & _
"                               sec_owner.OWNER_SECTION.op_sections(idx_op).OP_SECTION.stop + TIME_TOLERANCE > curr_time then" & vbLf & _
"                                for j : NUMERIC is size( sec_owner.OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections ) to 1 step -1 loop" & vbLf & _
"                                    exit when sec_owner.OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(j).EXP_SECTION.for_shimming = FALSE;" & vbLf & _
"                                    sec_owner.OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(j).EXP_SECTION.for_shimming := TRUE;" & vbLf & _
"                                end loop;" & vbLf & _
"                                sec_owner.OWNER_SECTION.op_sections(idx_op).OP_SECTION.did_shimming := TRUE;" & vbLf & _
"                            end if;" & vbLf & _
"                        elsif ""INFO : Instrument ownership ended for"" in line then" & vbLf & _
"                            if target_field = ""Owner Began"" or else curr_time >= start_time then" & vbLf & _
"                                sec_owner.OWNER_SECTION.stop := curr_time;" & vbLf & _
"                                owner_sections := owner_sections & {!sec_owner};" & vbLf & _
"                            end if;" & vbLf & _
"                            state := ""find ownership began"";" & vbLf & _
"                        end if;" & vbLf & _
"" & vbLf & _
"                    when ""find experiment start"" =>" & vbLf & _
"                        if ""INFO : Starting Experiment"" in line then" & vbLf & _
"                            sec_exp := new EXP_SECTION;" & vbLf & _
"                            sec_exp.EXP_SECTION.start        := curr_time;" & vbLf & _
"                            sec_exp.EXP_SECTION.for_shimming := sec_op.OP_SECTION.did_calibration;" & vbLf & _
"                            state := ""find experiment end"";" & vbLf & _
"                        elsif ""INFO : Starting Job"" in line then" & vbLf & _
"                            job_name := nth( line, POS_JOB_START, size( line ) - POS_JOB_START + 1 );" & vbLf & _
"                            job_name := nth( job_name, 1, index( ""'"", job_name ) - 1 );" & vbLf & _
"                            sec_op.OP_SECTION.job_name := job_name;" & vbLf & _
"                        elsif ""Gradient Shim: Start"" in line then" & vbLf & _
"                            sec_op.OP_SECTION.did_calibration := TRUE;" & vbLf & _
"                            sec_op.OP_SECTION.did_shimming    := TRUE;" & vbLf & _
"                        elsif ""INFO : Shimming..."" in line and then size( sec_op.OP_SECTION.exp_sections ) > 0 then" & vbLf & _
"                            idx := size( sec_op.OP_SECTION.exp_sections );" & vbLf & _
"                            sec_op.OP_SECTION.exp_sections(idx).EXP_SECTION.for_shimming := FALSE;" & vbLf & _
"                        elsif ""Terminating gradient shim.""        in line or else" & vbLf & _
"                              ""Gradient Shim: Completed in ""      in line or else" & vbLf & _
"                              ""CANCELLED : Aborted gradient shim"" in line then" & vbLf & _
"                            for j : NUMERIC is size( sec_op.OP_SECTION.exp_sections ) to 1 step -1 loop" & vbLf & _
"                                exit when sec_op.OP_SECTION.exp_sections(j).EXP_SECTION.for_shimming = FALSE;" & vbLf & _
"                                sec_op.OP_SECTION.exp_sections(j).EXP_SECTION.for_shimming := TRUE;" & vbLf & _
"                            end loop;" & vbLf & _
"                            sec_op.OP_SECTION.did_shimming := TRUE;" & vbLf & _
"                        elsif index( "" FATAL : "", line ) = POS_TIME_END + 1 then" & vbLf & _
"                            line := remove( line, 1, POS_TIME_END + 1 );" & vbLf & _
"                            if sec_op.OP_SECTION.first_fatal = NULL then" & vbLf & _
"                                sec_op.OP_SECTION.first_fatal := line;" & vbLf & _
"                            else" & vbLf & _
"                                sec_op.OP_SECTION.last_fatal  := line;" & vbLf & _
"                            end if;" & vbLf & _
"                        elsif ""INFO : Instrument operation ended for"" in line then" & vbLf & _
"                            sec_op.OP_SECTION.stop := curr_time;" & vbLf & _
"                            sec_owner.OWNER_SECTION.op_sections := sec_owner.OWNER_SECTION.op_sections & {sec_op};" & vbLf & _
"                            state := ""find operation began"";" & vbLf & _
"                        end if;" & vbLf & _
"" & vbLf & _
"                    when ""find experiment end"" =>" & vbLf & _
"                        if ""INFO : Post-experiment Default Initialization"" in line then" & vbLf & _
"                            sec_exp.EXP_SECTION.stop := curr_time;" & vbLf & _
"                            sec_op.OP_SECTION.exp_sections := sec_op.OP_SECTION.exp_sections & {sec_exp};" & vbLf & _
"                            state := ""find experiment start"";" & vbLf & _
"                        elsif index( "" FATAL : "", line ) = POS_TIME_END + 1 then" & vbLf & _
"                            line := remove( line, 1, POS_TIME_END + 1 );" & vbLf & _
"                            if sec_exp.EXP_SECTION.first_fatal = NULL then" & vbLf & _
"                                sec_exp.EXP_SECTION.first_fatal := line;" & vbLf & _
"                            else" & vbLf & _
"                                sec_exp.EXP_SECTION.last_fatal  := line;" & vbLf & _
"                            end if;" & vbLf & _
"                            if sec_op.OP_SECTION.first_fatal = NULL then" & vbLf & _
"                                sec_op.OP_SECTION.first_fatal := line;" & vbLf & _
"                            else" & vbLf & _
"                                sec_op.OP_SECTION.last_fatal  := line;" & vbLf & _
"                            end if;" & vbLf & _
"                        end if;" & vbLf & _
"                end case;" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"        if state = ""find operation began"" and then size( sec_owner ) > 0 then" & vbLf & _
"            sec_owner.OWNER_SECTION.stop := curr_time;" & vbLf & _
"            owner_sections := owner_sections & {!sec_owner};" & vbLf & _
"        end if;" & vbLf & _
"        sz := size( owner_sections );" & vbLf & _
"        if sz = 0 then" & vbLf & _
"            show_message( {""No owner section was detected."",""Please consider wider date range.""}, MSG_WARNING, ""enable calculate"" );" & vbLf & _
"            return;" & vbLf & _
"        elsif target_field = ""Operation Began"" then" & vbLf & _
"            if owner_sections(1).OWNER_SECTION.start < start_time and then" & vbLf & _
"               size( owner_sections(1).OWNER_SECTION.op_sections ) > 0 then" & vbLf & _
"                owner_sections(1).OWNER_SECTION.start := owner_sections(1).OWNER_SECTION.op_sections(1).OP_SECTION.start;" & vbLf & _
"            end if;" & vbLf & _
"            if owner_sections(sz).OWNER_SECTION.stop > stop_time and then" & vbLf & _
"               size( owner_sections(sz).OWNER_SECTION.op_sections ) > 0 then" & vbLf & _
"                idx := size( owner_sections(sz).OWNER_SECTION.op_sections );" & vbLf & _
"                owner_sections(sz).OWNER_SECTION.stop := owner_sections(sz).OWNER_SECTION.op_sections(idx).OP_SECTION.stop;" & vbLf & _
"            end if;" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        -- parse queue_history.log file" & vbLf & _
"        rate      := 0;" & vbLf & _
"        idx_op    := 1;" & vbLf & _
"        idx_owner := 1;" & vbLf & _
"        files     := ((logs(""queue_history-old.log"") = NULL) ? {} : {""queue_history-old.log""}) & {""queue_history.log""};" & vbLf & _
"        steps     := {100 / size( files ), NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now parsing queue_history.log file ..."" );" & vbLf & _
"" & vbLf & _
"        for fl in files loop" & vbLf & _
"            init_line_info( logs(fl) );" & vbLf & _
"            steps(2) := steps(1) / line_count;" & vbLf & _
"" & vbLf & _
"            for i : NUMERIC is 1 to line_count loop" & vbLf & _
"                rate := rate + steps(2);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"                line := get_next_line;" & vbLf & _
"" & vbLf & _
"                -- get job_date, job_name and job_result" & vbLf & _
"                --   27-JAN-2017 15:49:23 New Job 1 : FINISHED" & vbLf & _
"                --   31-JAN-2017 10:48:34 BDISnSe2PCy2 (1) : CANCELLED" & vbLf & _
"                --   20-JAN-2018 19:45:24 INFO : qBird 5 : FINISHED" & vbLf & _
"                --   15-JAN-2018 14:48:08 INFO : test : CANCELLED : Failed to load sample in slot 10." & vbLf & _
"                job_date   :=                    date( nth( line, POS_LINE_START, POS_TIME_END ) );" & vbLf & _
"                line       :=                       remove( line, POS_LINE_START, POS_TIME_END + 1 );" & vbLf & _
"                line       := (""INFO : "" in line) ? remove( line, POS_LINE_START, size( ""INFO : "" ) ) : line;" & vbLf & _
"                job_name   :=                          nth( line, 1, index( "" : "", line )                 - 1 );" & vbLf & _
"                job_result :=                       remove( line, 1, index( "" : "", line ) + size( "" : "" ) - 1 );" & vbLf & _
"                job_result := ("" : "" in job_result) ?  nth( job_result, 1, index( "" : "", job_result ) - 1) : job_result;" & vbLf & _
"" & vbLf & _
"                loop" & vbLf & _
"                    if idx_owner > size( owner_sections ) then" & vbLf & _
"                        exit;" & vbLf & _
"                    elsif idx_op > size( owner_sections(idx_owner).OWNER_SECTION.op_sections ) then" & vbLf & _
"                        idx_owner := idx_owner + 1;" & vbLf & _
"                        idx_op    := 1;" & vbLf & _
"                        next;" & vbLf & _
"                    end if;" & vbLf & _
"                    op_stop := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.stop;" & vbLf & _
"                    if job_date < op_stop then" & vbLf & _
"                        exit;" & vbLf & _
"                    elsif job_date - op_stop < JOB_TIME_TOLERANCE then" & vbLf & _
"                        owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.job_name   := job_name;" & vbLf & _
"                        owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.job_result := job_result;" & vbLf & _
"                        idx_op := idx_op + 1;" & vbLf & _
"                        exit;" & vbLf & _
"                    else" & vbLf & _
"                        idx_op := idx_op + 1;" & vbLf & _
"                    end if;" & vbLf & _
"                end loop;" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        -- query name_id of EXPERIMENT parameter" & vbLf & _
"        query_text := ""SELECT id FROM parameter_names WHERE name = 'EXPERIMENT'"";" & vbLf & _
"        result := query( nc, query_text);" & vbLf & _
"        if result = SC~F~CONNECT_FAILED then" & vbLf & _
"            show_message( ""Spectrometer was disconnected."", MSG_WARNING, ""disable connect"" );" & vbLf & _
"            next_action := ""connect"";" & vbLf & _
"            return;" & vbLf & _
"        end if;" & vbLf & _
"        name_id := result(1,1);" & vbLf & _
"" & vbLf & _
"        -- query data ID" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( owner_sections );" & vbLf & _
"        steps := {100 / sz, NULL, NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now querying data ID ..."" );" & vbLf & _
"" & vbLf & _
"        for idx_owner is 1 to sz loop" & vbLf & _
"            sz       := size( owner_sections(idx_owner).OWNER_SECTION.op_sections );" & vbLf & _
"            rate     := (sz = 0) ? (rate + steps(1)) : rate;" & vbLf & _
"            steps(2) := (sz = 0) ? NULL : (steps(1) / sz);" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            for idx_op is 1 to sz loop" & vbLf & _
"                sz       := size( owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections );" & vbLf & _
"                rate     := (sz = 0) ? (rate + steps(2)) : rate;" & vbLf & _
"                steps(3) := (sz = 0) ? NULL : (steps(2) / sz);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                for idx_exp is 1 to sz loop" & vbLf & _
"                    rate := rate + steps(3);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                    exp_start := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.start;" & vbLf & _
"                    exp_stop  := (idx_exp < sz) ? owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp+1).EXP_SECTION.start :" & vbLf & _
"                                                  owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.stop;" & vbLf & _
"" & vbLf & _
"                    query_text := ""SELECT id FROM file_table WHERE (creation_time >= "" & [STRING](exp_start^[s]^[]) & "") AND (creation_time <= "" & [STRING](exp_stop ^[s]^[]) & "") ORDER BY id ASC"";" & vbLf & _
"                    result := query( nc, query_text);" & vbLf & _
"                    if result = SC~F~CONNECT_FAILED then" & vbLf & _
"                        show_message( ""Spectrometer was disconnected."", MSG_WARNING, ""disable connect"" );" & vbLf & _
"                        next_action := ""connect"";" & vbLf & _
"                        return;" & vbLf & _
"                    elsif result = {} then" & vbLf & _
"                        next;" & vbLf & _
"                    end if;" & vbLf & _
"                    data_id := result(1,1);" & vbLf & _
"" & vbLf & _
"                    owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.data_id := data_id;" & vbLf & _
"                    idxs_list := idxs_list & {{idx_owner,idx_op,idx_exp}};" & vbLf & _
"                end loop;" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        -- query data information" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( idxs_list );" & vbLf & _
"        steps := {100 / sz};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now querying data information ..."" );" & vbLf & _
"" & vbLf & _
"        for i : NUMERIC is 1 to sz step MAX_REC_SIZE loop" & vbLf & _
"            rate := steps(1) * i;" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            query_text := """";" & vbLf & _
"            curr_list  := nth( idxs_list, i, MAX_REC_SIZE );" & vbLf & _
"            for pos in curr_list loop" & vbLf & _
"                idx_owner  := pos(1);" & vbLf & _
"                idx_op     := pos(2);" & vbLf & _
"                idx_exp    := pos(3);" & vbLf & _
"                query_text := query_text & ((query_text = """") ? """" : "", "") &" & vbLf & _
"                    owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.data_id;" & vbLf & _
"            end loop;" & vbLf & _
"" & vbLf & _
"            query_text := use_folder ? (""SELECT file.id, file.name, param.value, file.project_id, proj.name, folder.name FROM file_table file LEFT JOIN parameter_values param ON file.id = param.file_id AND param.name_id = "" & name_id & "" LEFT JOIN project_table proj ON file.project_id = proj.id LEFT JOIN folder_table folder ON file.id = folder.file_id WHERE file.id IN( "" & query_text & "" ) ORDER BY file.id ASC, param.value DESC"") :" & vbLf & _
"                                       (""SELECT file.id, file.name, param.value, file.project_id, proj.name              FROM file_table file LEFT JOIN parameter_values param ON file.id = param.file_id AND param.name_id = "" & name_id & "" LEFT JOIN project_table proj ON file.project_id = proj.id                                                           WHERE file.id IN( "" & query_text & "" ) ORDER BY file.id ASC, param.value DESC"");" & vbLf & _
"            result := query( nc, query_text);" & vbLf & _
"            if result = SC~F~CONNECT_FAILED then" & vbLf & _
"                show_message( ""Spectrometer was disconnected."", MSG_WARNING, ""disable connect"" );" & vbLf & _
"                next_action := ""connect"";" & vbLf & _
"                return;" & vbLf & _
"            end if;" & vbLf & _
"" & vbLf & _
"            sz  := size( result );" & vbLf & _
"            idx := 0;" & vbLf & _
"            for pos in curr_list loop" & vbLf & _
"                idx_owner  := pos(1);" & vbLf & _
"                idx_op     := pos(2);" & vbLf & _
"                idx_exp    := pos(3);" & vbLf & _
"                data_id    := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.data_id;" & vbLf & _
"" & vbLf & _
"                loop" & vbLf & _
"                    idx := idx + 1;" & vbLf & _
"                    exit when idx > sz;" & vbLf & _
"                    if result(idx).QUERY_DATA.data_id = data_id then" & vbLf & _
"                        owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.data_file  := result(idx).QUERY_DATA.data_file;" & vbLf & _
"                        owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.exp_file   := result(idx).QUERY_DATA.exp_file;" & vbLf & _
"                        if result(idx).QUERY_DATA.project_id /= ""0"" and then" & vbLf & _
"                           owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.project_id = NULL then" & vbLf & _
"                            owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.project_id                               := result(idx).QUERY_DATA.project_id;" & vbLf & _
"                            owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.project_name                             := result(idx).QUERY_DATA.project_name;" & vbLf & _
"                        end if;" & vbLf & _
"                        if use_folder and then result(idx).QUERY_DATA.folder /= ""NULL"" then" & vbLf & _
"                            owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.folder := result(idx).QUERY_DATA.folder;" & vbLf & _
"                            folder_used := TRUE;" & vbLf & _
"                        end if;" & vbLf & _
"                        exit;" & vbLf & _
"                    end if;" & vbLf & _
"                end loop;" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        -- summarize data" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( owner_sections );" & vbLf & _
"        steps := {100 / sz, NULL, NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now summarizing data ..."" );" & vbLf & _
"" & vbLf & _
"        for idx_owner is 1 to sz loop" & vbLf & _
"            sz       := size( owner_sections(idx_owner).OWNER_SECTION.op_sections );" & vbLf & _
"            rate     := (sz = 0) ? (rate + steps(1)) : rate;" & vbLf & _
"            steps(2) := (sz = 0) ? NULL : (steps(1) / sz);" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            username := owner_sections(idx_owner).OWNER_SECTION.username;" & vbLf & _
"            idx_user := index( username, user_list );" & vbLf & _
"            if idx_user = 0 then" & vbLf & _
"                data_user := new USER_DATA;" & vbLf & _
"                data_user.USER_DATA.username   := username;" & vbLf & _
"                data_user.USER_DATA.owner_time := 0[s];" & vbLf & _
"                data_user.USER_DATA.op_count   := 0;" & vbLf & _
"                data_user.USER_DATA.exp_count  := 0;" & vbLf & _
"                data_user.USER_DATA.owners     := {};" & vbLf & _
"                data_by_user := data_by_user & {data_user};" & vbLf & _
"                user_list    := user_list & {username};" & vbLf & _
"                idx_user     := size( data_by_user );" & vbLf & _
"            end if;" & vbLf & _
"" & vbLf & _
"            data_owner := new OWNER_DATA;" & vbLf & _
"            data_owner.OWNER_DATA.start := owner_sections(idx_owner).OWNER_SECTION.start;" & vbLf & _
"            data_owner.OWNER_DATA.stop  := owner_sections(idx_owner).OWNER_SECTION.stop;" & vbLf & _
"            data_owner.OWNER_DATA.ops   := repeat( {}, -sz );" & vbLf & _
"" & vbLf & _
"            for idx_op is 1 to sz loop" & vbLf & _
"                sz       := size( owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections );" & vbLf & _
"                rate     := (sz = 0) ? (rate + steps(2)) : rate;" & vbLf & _
"                steps(3) := (sz = 0) ? NULL : (steps(2) / sz);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                proj_name := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.project_name;" & vbLf & _
"                idx_proj  := index( proj_name, proj_list );" & vbLf & _
"                if idx_proj = 0 then" & vbLf & _
"                    data_proj := new PROJECT_DATA;" & vbLf & _
"                    data_proj.PROJECT_DATA.project_name := proj_name;" & vbLf & _
"                    data_proj.PROJECT_DATA.op_time      := 0[s];" & vbLf & _
"                    data_proj.PROJECT_DATA.exp_count    := 0;" & vbLf & _
"                    data_proj.PROJECT_DATA.ops          := {};" & vbLf & _
"                    data_by_proj := data_by_proj & {data_proj};" & vbLf & _
"                    proj_list    := proj_list & {proj_name};" & vbLf & _
"                    idx_proj     := size( data_by_proj );" & vbLf & _
"                end if;" & vbLf & _
"" & vbLf & _
"                data_op := new OP_DATA;" & vbLf & _
"                data_op.OP_DATA.username     := owner_sections(idx_owner).OWNER_SECTION.username;" & vbLf & _
"                data_op.OP_DATA.start        := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.start;" & vbLf & _
"                data_op.OP_DATA.stop         := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.stop;" & vbLf & _
"                data_op.OP_DATA.job_name     := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.job_name;" & vbLf & _
"                data_op.OP_DATA.job_result   := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.job_result;" & vbLf & _
"                data_op.OP_DATA.first_fatal  := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.first_fatal;" & vbLf & _
"                data_op.OP_DATA.project_name := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.project_name;" & vbLf & _
"                data_op.OP_DATA.exps         := repeat( {}, -sz );" & vbLf & _
"" & vbLf & _
"                for idx_exp is 1 to sz loop" & vbLf & _
"                    rate := rate + steps(3);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                    -- ignore sawtooth or FG shimming experiments" & vbLf & _
"                    next when owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.exp_file in {""sawtooth.jxp"", ""sawtooth_5001.jxp""} or else" & vbLf & _
"                              owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.for_shimming = TRUE;" & vbLf & _
"" & vbLf & _
"                    data_exp := new EXP_DATA;" & vbLf & _
"                    data_exp.EXP_DATA.folder    := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.folder;" & vbLf & _
"                    data_exp.EXP_DATA.exp_file  := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.exp_file;" & vbLf & _
"                    data_exp.EXP_DATA.data_file := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(idx_exp).EXP_SECTION.data_file;" & vbLf & _
"                    data_op.OP_DATA.exps := data_op.OP_DATA.exps & {data_exp};" & vbLf & _
"" & vbLf & _
"                    folder     := data_exp.EXP_DATA.folder;" & vbLf & _
"                    idx_folder := index( folder, folder_list );" & vbLf & _
"                    if idx_folder = 0 then" & vbLf & _
"                        data_folder := new FOLDER_DATA;" & vbLf & _
"                        data_folder.FOLDER_DATA.folder    := folder;" & vbLf & _
"                        data_folder.FOLDER_DATA.op_time   := 0[s];" & vbLf & _
"                        data_folder.FOLDER_DATA.exp_count := 0;" & vbLf & _
"                        data_folder.FOLDER_DATA.ops       := {};" & vbLf & _
"                        data_by_folder := data_by_folder & {data_folder};" & vbLf & _
"                        folder_list    := folder_list & {folder};" & vbLf & _
"                        start_list     := start_list & {{}};" & vbLf & _
"                        idx_folder     := size( data_by_folder );" & vbLf & _
"                    end if;" & vbLf & _
"" & vbLf & _
"                    idx := index( data_op.OP_DATA.start, start_list(idx_folder) );" & vbLf & _
"                    if idx = 0 then" & vbLf & _
"                        data := data_op;" & vbLf & _
"                        data.OP_DATA.exps := {};" & vbLf & _
"                        data_by_folder(idx_folder).FOLDER_DATA.ops     := data_by_folder(idx_folder).FOLDER_DATA.ops & {data};" & vbLf & _
"                        data_by_folder(idx_folder).FOLDER_DATA.op_time := data_by_folder(idx_folder).FOLDER_DATA.op_time + data_op.OP_DATA.stop - data_op.OP_DATA.start;" & vbLf & _
"                        start_list(idx_folder) := start_list(idx_folder) & {data.OP_DATA.start};" & vbLf & _
"                        idx := size( data_by_folder(idx_folder).FOLDER_DATA.ops );" & vbLf & _
"                    end if;" & vbLf & _
"                    data_by_folder(idx_folder).FOLDER_DATA.ops(idx).OP_DATA.exps := data_by_folder(idx_folder).FOLDER_DATA.ops(idx).OP_DATA.exps & {data_exp};" & vbLf & _
"                    data_by_folder(idx_folder).FOLDER_DATA.exp_count             := data_by_folder(idx_folder).FOLDER_DATA.exp_count + 1;" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"                -- add failed experiment title if first_fatal/last_fatal is like as follow" & vbLf & _
"                --   FATAL : Unable to collect data from Experiment : single_pulse" & vbLf & _
"                --   FATAL : Unable to find or open the experiment file (...\single_pulse.jxp) for Experiment 'single_pulse'" & vbLf & _
"                if size( data_op.OP_DATA.exps ) = 0 then" & vbLf & _
"                    exp_file := NULL;" & vbLf & _
"                    if ""FATAL : Unable to collect data from Experiment :"" in owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.last_fatal then" & vbLf & _
"                        line := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.last_fatal;" & vbLf & _
"                        exp_file := remove( line, 1, POS_FATAL_END );" & vbLf & _
"                    elsif ""FATAL : Unable to find or open the experiment file ("" in owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.first_fatal and then" & vbLf & _
"                          "") for Experiment "" in owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.first_fatal then" & vbLf & _
"                        line := owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.first_fatal;" & vbLf & _
"                        idx1 := index( ""("", line ) + 1;" & vbLf & _
"                        idx2 := index( "")"", line ) - 1;" & vbLf & _
"                        exp_file := nth( line, idx1, idx2 - idx1 + 1 );" & vbLf & _
"                    end if;" & vbLf & _
"                    if exp_file /= NULL then" & vbLf & _
"                        data_exp := new EXP_DATA;" & vbLf & _
"                        data_exp.EXP_DATA.exp_file := exp_file;" & vbLf & _
"                        data_op.OP_DATA.exps := data_op.OP_DATA.exps & {data_exp};" & vbLf & _
"                    end if;" & vbLf & _
"                end if;" & vbLf & _
"" & vbLf & _
"                -- ignore sawtooth or FG shimming jobs" & vbLf & _
"                next when        owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.job_name = ""Sawtooth"" or else" & vbLf & _
"                          (size( owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections ) = 1 and then" & vbLf & _
"                                 owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.exp_sections(1).EXP_SECTION.exp_file in {""sawtooth.jxp"", ""sawtooth_5001.jxp""}) or else" & vbLf & _
"                          (size( data_op.OP_DATA.exps ) = 0 and then" & vbLf & _
"                                 owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.did_shimming = TRUE);" & vbLf & _
"" & vbLf & _
"                -- handle no job_result" & vbLf & _
"                if data_op.OP_DATA.job_result   = NULL and then" & vbLf & _
"                   data_op.OP_DATA.first_fatal /= NULL then" & vbLf & _
"                    data_op.OP_DATA.job_result := ""FAILED"";" & vbLf & _
"                end if;" & vbLf & _
"" & vbLf & _
"                data_owner.OWNER_DATA.ops                     := data_owner.OWNER_DATA.ops & {data_op};" & vbLf & _
"                data_by_user(idx_user).USER_DATA.exp_count    := data_by_user(idx_user).USER_DATA.exp_count + size( data_op.OP_DATA.exps );" & vbLf & _
"                data_by_user(idx_user).USER_DATA.op_count     := data_by_user(idx_user).USER_DATA.op_count + 1;" & vbLf & _
"" & vbLf & _
"                -- delete EXP_DATA with no data_file for PROJECT_DATA" & vbLf & _
"                for i : NUMERIC is size( data_op.OP_DATA.exps ) to 1 step -1 loop" & vbLf & _
"                    if data_op.OP_DATA.exps(i).EXP_DATA.data_file = NULL then" & vbLf & _
"                        data_op.OP_DATA.exps := remove( data_op.OP_DATA.exps, i );" & vbLf & _
"                    end if;" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"                -- delete PROJECT_DATA with no OP_DATA" & vbLf & _
"                if size( data_op.OP_DATA.exps ) = 0 then" & vbLf & _
"                    if size( data_by_proj(idx_proj).PROJECT_DATA.ops ) = 0 then" & vbLf & _
"                        proj_list    := remove( proj_list,    idx_proj );" & vbLf & _
"                        data_by_proj := remove( data_by_proj, idx_proj );" & vbLf & _
"                    end if;" & vbLf & _
"                else" & vbLf & _
"                    data_by_proj(idx_proj).PROJECT_DATA.ops       := data_by_proj(idx_proj).PROJECT_DATA.ops & {data_op};" & vbLf & _
"                    data_by_proj(idx_proj).PROJECT_DATA.exp_count := data_by_proj(idx_proj).PROJECT_DATA.exp_count + size( data_op.OP_DATA.exps );" & vbLf & _
"                    data_by_proj(idx_proj).PROJECT_DATA.op_time   := data_by_proj(idx_proj).PROJECT_DATA.op_time +" & vbLf & _
"                                                                         owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.stop -" & vbLf & _
"                                                                         owner_sections(idx_owner).OWNER_SECTION.op_sections(idx_op).OP_SECTION.start;" & vbLf & _
"                end if;" & vbLf & _
"            end loop;" & vbLf & _
"            data_by_user(idx_user).USER_DATA.owners     := data_by_user(idx_user).USER_DATA.owners & {data_owner};" & vbLf & _
"            data_by_user(idx_user).USER_DATA.owner_time := data_by_user(idx_user).USER_DATA.owner_time +" & vbLf & _
"                                                               owner_sections(idx_owner).OWNER_SECTION.stop -" & vbLf & _
"                                                               owner_sections(idx_owner).OWNER_SECTION.start;" & vbLf & _
"        end loop;" & vbLf & _
"        data_by_user   := sort( data_by_user   );" & vbLf & _
"        data_by_proj   := sort( data_by_proj   );" & vbLf & _
"        data_by_folder := sort( data_by_folder );" & vbLf & _
"" & vbLf & _
"        -- build table by username" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( data_by_user );" & vbLf & _
"        steps := {100 / sz, NULL, NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now building table by username ..."" );" & vbLf & _
"" & vbLf & _
"        for i : NUMERIC is 1 to sz loop" & vbLf & _
"            rate := rate + steps(1);" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            row := new USER_TABLE_ROW;" & vbLf & _
"            row.USER_TABLE_ROW.username    :=                     data_by_user(i).USER_DATA.username;" & vbLf & _
"            row.USER_TABLE_ROW.owner_time  := to_duration_string( data_by_user(i).USER_DATA.owner_time );" & vbLf & _
"            row.USER_TABLE_ROW.owner_count :=               size( data_by_user(i).USER_DATA.owners );" & vbLf & _
"            row.USER_TABLE_ROW.op_count    :=                     data_by_user(i).USER_DATA.op_count;" & vbLf & _
"            row.USER_TABLE_ROW.exp_count   :=                     data_by_user(i).USER_DATA.exp_count;" & vbLf & _
"" & vbLf & _
"            table_by_user := table_by_user & {row};" & vbLf & _
"        end loop;" & vbLf & _
"        table_by_user := {{WM~SET_VALUE, table_by_user}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_ROW_HEIGHT, 1..sz, 2}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_COLUMN_HEADING,   {""Username"", ""Owner Time"", ""Owner Count"", ""Jobs"",   ""Experiments""}}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_COLUMN_ALIGNMENT, {""center"",   ""center"",     ""center"",      ""center"", ""center""     }}," & vbLf & _
"                          {WM~SET_FEATURE, WM~NO_DRAW, FALSE}};" & vbLf & _
"" & vbLf & _
"        -- build table by project name" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( data_by_proj );" & vbLf & _
"        steps := {100 / sz, NULL, NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now building table by project name ..."" );" & vbLf & _
"" & vbLf & _
"        for i : NUMERIC is 1 to sz loop" & vbLf & _
"            rate := rate + steps(1);" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            row := new PROJECT_TABLE_ROW;" & vbLf & _
"            row.PROJECT_TABLE_ROW.project_name :=          to_string( data_by_proj(i).PROJECT_DATA.project_name, ""(No Project Name)"" );" & vbLf & _
"            row.PROJECT_TABLE_ROW.op_time      := to_duration_string( data_by_proj(i).PROJECT_DATA.op_time );" & vbLf & _
"            row.PROJECT_TABLE_ROW.op_count     :=               size( data_by_proj(i).PROJECT_DATA.ops );" & vbLf & _
"            row.PROJECT_TABLE_ROW.exp_count    :=                     data_by_proj(i).PROJECT_DATA.exp_count;" & vbLf & _
"" & vbLf & _
"            table_by_proj := table_by_proj & {row};" & vbLf & _
"        end loop;" & vbLf & _
"        table_by_proj := {{WM~SET_VALUE, table_by_proj}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_ROW_HEIGHT, 1..sz, 2}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_COLUMN_HEADING,   {""Project Name"", ""Operation Time"", ""Jobs"",   ""Experiments""}}," & vbLf & _
"                          {WM~SET_VALUE, WM~SET_COLUMN_ALIGNMENT, {""center"",       ""center"",         ""center"", ""center""     }}," & vbLf & _
"                          {WM~SET_FEATURE, WM~NO_DRAW, FALSE}};" & vbLf & _
"" & vbLf & _
"        -- build table by folder name" & vbLf & _
"        rate  := 0;" & vbLf & _
"        sz    := size( data_by_folder );" & vbLf & _
"        steps := {100 / sz, NULL, NULL};" & vbLf & _
"        update_indicator( ""update"", 0, ""Now building table by folder name ..."" );" & vbLf & _
"" & vbLf & _
"        for i : NUMERIC is 1 to sz loop" & vbLf & _
"            rate := rate + steps(1);" & vbLf & _
"            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"            row := new FOLDER_TABLE_ROW;" & vbLf & _
"            row.FOLDER_TABLE_ROW.folder    :=          to_string( data_by_folder(i).FOLDER_DATA.folder, ""(No Folder Name)"" );" & vbLf & _
"            row.FOLDER_TABLE_ROW.op_time   := to_duration_string( data_by_folder(i).FOLDER_DATA.op_time );" & vbLf & _
"            row.FOLDER_TABLE_ROW.op_count  :=               size( data_by_folder(i).FOLDER_DATA.ops );" & vbLf & _
"            row.FOLDER_TABLE_ROW.exp_count :=                     data_by_folder(i).FOLDER_DATA.exp_count;" & vbLf & _
"" & vbLf & _
"            table_by_folder := table_by_folder & {row};" & vbLf & _
"        end loop;" & vbLf & _
"        table_by_folder := {{WM~SET_VALUE, table_by_folder}," & vbLf & _
"                            {WM~SET_VALUE, WM~SET_ROW_HEIGHT, 1..sz, 2}," & vbLf & _
"                            {WM~SET_VALUE, WM~SET_COLUMN_HEADING,   {""Folder Name"", ""Operation Time"", ""Jobs"",   ""Experiments""}}," & vbLf & _
"                            {WM~SET_VALUE, WM~SET_COLUMN_ALIGNMENT, {""center"",      ""center"",         ""center"", ""center""     }}," & vbLf & _
"                            {WM~SET_FEATURE, WM~NO_DRAW, FALSE}};" & vbLf & _
"" & vbLf & _
"        -- enable calculation" & vbLf & _
"        conf := {start_date, stop_date, target_field};" & vbLf & _
"        log_message( ""Summarized machine time data successfully."", MSG_INFO );" & vbLf & _
"        show_message( ""Please click [Save] icon to save current machine time data into CSV file."", MSG_INFO, ""enable save"" );" & vbLf & _
"    end CALCULATE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define HANDLE_ACTION is" & vbLf & _
"        case next_action is" & vbLf & _
"            when ""connect"" =>" & vbLf & _
"                connect;" & vbLf & _
"" & vbLf & _
"            when ""calculate"" =>" & vbLf & _
"                calculate;" & vbLf & _
"        end case;" & vbLf & _
"    end HANDLE_ACTION;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function HANDLE_CALENDAR( pname : STRING, pvalue : WILDCARD ) is" & vbLf & _
"        var" & vbLf & _
"            cal_date : NUMERIC," & vbLf & _
"            showp    : BOOLEAN := to_boolean( pvalue )," & vbLf & _
"            is_start : BOOLEAN := pname = ""start date"";" & vbLf & _
"" & vbLf & _
"        if showp then" & vbLf & _
"            if cal_target /= """" then" & vbLf & _
"                sw {{def,is_start ? ""tb_stop_date"":""tb_start_date""}, {WM~SET_STATE, WM~BUTTON_OFF}};" & vbLf & _
"            end if;" & vbLf & _
"            cal_target := pname;" & vbLf & _
"            cal_id     := (cal_id = """") ? create_calendar_window( def, ""fr_main"" ) : cal_id;" & vbLf & _
"            cal_date   := [NUMERIC](to_date_time( is_start ? start_date : stop_date ));" & vbLf & _
"            cal_pos    := {WM~LEFT_BOTTOM_RIGHT_TOP, -55[mm],NULL,9[mm],is_start ? 9[mm]:18[mm]};" & vbLf & _
"            set_calendar( cal_id, cal_date, cal_pos );" & vbLf & _
"        elsif pname = cal_target then" & vbLf & _
"            cal_target := """";" & vbLf & _
"            hide_calendar( cal_id );" & vbLf & _
"        end if;" & vbLf & _
"" & vbLf & _
"        sw {{def,""pb_action""}, {WM~SET_FEATURE, WM~ENABLE, cal_target = """"}};" & vbLf & _
"    end HANDLE_CALENDAR;" & vbLf & _
"" & vbLf & _
"    ----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function HANDLE_CALENDAR_MONTH( cal_id : STRING, months : NUMERIC ) is" & vbLf & _
"        set_calendar_month( cal_id, months, cal_pos );" & vbLf & _
"    end HANDLE_CALENDAR_MONTH;" & vbLf & _
"" & vbLf & _
"    ----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function HANDLE_CALENDAR_SELECT( cal_id : STRING, day : NUMERIC ) is" & vbLf & _
"        var" & vbLf & _
"            is_start : BOOLEAN := cal_target = ""start date""," & vbLf & _
"            ib_wdg   : STRING  := is_start ? ""ib_start_date"" : ""ib_stop_date""," & vbLf & _
"            tb_wdg   : STRING  := is_start ? ""tb_start_date"" : ""tb_stop_date""," & vbLf & _
"            date_str : STRING  := to_date_string( calendar_year, calendar_month, day );" & vbLf & _
"" & vbLf & _
"        sw {def, {ib_wdg, {WM~SET_VALUE, date_str}}," & vbLf & _
"                 {tb_wdg, {WM~SET_STATE, WM~BUTTON_OFF}}};" & vbLf & _
"        sw {WM~SEND_ADDRESS, {def,ib_wdg}, {WM~GET_SELECTED}};" & vbLf & _
"    end HANDLE_CALENDAR_SELECT;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define HANDLE_UPDATE_SPECS is" & vbLf & _
"        update_specs;" & vbLf & _
"    end HANDLE_UPDATE_SPECS;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define HANDLE_SAVE is" & vbLf & _
"        var" & vbLf & _
"            UPDATE_INTERVAL : constant UNIVERSAL := 0.5[s]," & vbLf & _
"            DAY             : constant UNIVERSAL := 24 * 60 * 60[s]," & vbLf & _
"" & vbLf & _
"            line        : SET," & vbLf & _
"            text        : SET," & vbLf & _
"            order       : SET," & vbLf & _
"            steps       : SET," & vbLf & _
"            selected    : SET," & vbLf & _
"            msg         : STRING," & vbLf & _
"            spec        : STRING," & vbLf & _
"            group       : STRING," & vbLf & _
"            rep_str     : STRING," & vbLf & _
"            csv_file    : STRING," & vbLf & _
"            match_str   : STRING," & vbLf & _
"            pretty_name : STRING," & vbLf & _
"            sz          : NUMERIC," & vbLf & _
"            rate        : NUMERIC," & vbLf & _
"            result      : WILDCARD," & vbLf & _
"            prev        : UNIVERSAL," & vbLf & _
"            duration    : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        -- get status" & vbLf & _
"        spec     :=  cw {{def,""se_spec"" }, {WM~GET_SELECTED}};" & vbLf & _
"        group    := TAB_NAMES(cw {{def,""tabs""}, {WM~GET_SELECTED}});" & vbLf & _
"        order    :=  cw {{def,""tbl_data""}, {WM~GET_INDEX}};" & vbLf & _
"        selected := (cw {{def,""tbl_data""}, {WM~GET_SELECTED, WM~GET_LOCATION, WM~Y}})(3);" & vbLf & _
"        selected := (selected = {}) ? order : selected;" & vbLf & _
"" & vbLf & _
"        -- get CSV filename" & vbLf & _
"        match_str := is_jpn ? ""/"" : ""-"";" & vbLf & _
"        rep_str   := is_jpn ? ""_"" : ""_"";" & vbLf & _
"        csv_file := ""machine_time-"" &" & vbLf & _
"                    lower( spec ) & ""-"" &" & vbLf & _
"                    replace_string( start_date, match_str, rep_str ) & ""-"" &" & vbLf & _
"                    replace_string(  stop_date, match_str, rep_str ) & ""-"" &" & vbLf & _
"                    ((group = ""Username"") ? ""username"" : ((group = ""Project Name"") ? ""project_name"" : ""folder_name"")) & "".csv"";" & vbLf & _
"        pretty_name := ""machine time CSV file for "" & lower( group );" & vbLf & _
"        csv_file    := file_dialog( TL_SAVE_FILE, ""Save "" & pretty_name & "" as..."", INT_REPORT, ""*.csv"", csv_file );" & vbLf & _
"        return when type csv_file /= [STRING];" & vbLf & _
"" & vbLf & _
"        -- build CSV text" & vbLf & _
"        rate  := 0;" & vbLf & _
"        prev  := seconds;" & vbLf & _
"        steps := {NULL,NULL,NULL,NULL};" & vbLf & _
"        handle_state( ""enable abort"", ""Saving "" & pretty_name & "" was cancelled."", ""enable save"" );" & vbLf & _
"        update_indicator( ""update"", 0, ""Now making "" & pretty_name & "" ..."" );" & vbLf & _
"        text  := {""* Summary""," & vbLf & _
"                  {""Spectrometer"",         spec}," & vbLf & _
"                  {""Grouped by"",          group}," & vbLf & _
"                  {""Target Field"", target_field}," & vbLf & _
"                  {""From:"",          start_date}," & vbLf & _
"                  {""To:"",             stop_date}," & vbLf & _
"                  """"};" & vbLf & _
"        case group is" & vbLf & _
"            when ""Username"" =>" & vbLf & _
"                text := text & {{""Username"",""Owner Time"",""Owner Time/day"",""Owner Count"",""Jobs"",""Experiments""}};" & vbLf & _
"" & vbLf & _
"                sz       := size( order );" & vbLf & _
"                rate     := (sz = 0) ? (rate + 10) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (10 / sz);" & vbLf & _
"" & vbLf & _
"                for idx_user in order loop" & vbLf & _
"                    rate := rate + steps(1);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                    line := repeat( {""""}, rank.USER_SUMMARY_FORMAT );" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.username    :=                     data_by_user(idx_user).USER_DATA.username;" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.owner_time  := to_duration_string( data_by_user(idx_user).USER_DATA.owner_time );" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.owner_days  :=                     data_by_user(idx_user).USER_DATA.owner_time / DAY;" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.owner_count :=               size( data_by_user(idx_user).USER_DATA.owners );" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.op_count    :=                     data_by_user(idx_user).USER_DATA.op_count;" & vbLf & _
"                    line.USER_SUMMARY_FORMAT.exp_count   :=                     data_by_user(idx_user).USER_DATA.exp_count;" & vbLf & _
"                    text := text & {line};" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"                rate     := (sz = 0) ? (rate + 90) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (90 / sz);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                for idx_user in order loop" & vbLf & _
"                    sz       := size( data_by_user(idx_user).USER_DATA.owners );" & vbLf & _
"                    rate     := (sz = 0 or else idx_user not in selected) ? (rate + steps(1)) : rate;" & vbLf & _
"                    steps(2) := (sz = 0) ? NULL : (steps(1) / sz);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"                    next when idx_user not in selected;" & vbLf & _
"" & vbLf & _
"                    line := {""Owner Began"",""Owner End"",""Duration"",""Duration/day"",""Jobs"",""Experiments"",""Job Name"",""Experiment Name"",""Project Name"",""Folder"",""Filename"",""Job Result"",""Note""};" & vbLf & _
"                    line := folder_used ? line : remove( line, rank.USER_FORMAT.folder );" & vbLf & _
"                    text := text & {""""," & vbLf & _
"                                    ""* Breakdown of user '"" & data_by_user(idx_user).USER_DATA.username & ""'""," & vbLf & _
"                                    line};" & vbLf & _
"" & vbLf & _
"                    for idx_owner : NUMERIC is 1 to sz loop" & vbLf & _
"                        sz       := size( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops );" & vbLf & _
"                        rate     := (sz = 0) ? (rate + steps(2)) : rate;" & vbLf & _
"                        steps(3) := (sz = 0) ? NULL : (steps(2) / sz);" & vbLf & _
"                        return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                        duration := data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.stop -" & vbLf & _
"                                    data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.start;" & vbLf & _
"" & vbLf & _
"                        line := repeat( {""""}, rank.USER_FORMAT );" & vbLf & _
"                        line.USER_FORMAT.owner_start := to_date_time_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.start );" & vbLf & _
"                        line.USER_FORMAT.owner_stop  := to_date_time_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.stop  );" & vbLf & _
"                        line.USER_FORMAT.job_count   := sz;" & vbLf & _
"                        line.USER_FORMAT.exp_count   := 0;" & vbLf & _
"                        line.USER_FORMAT.duration    := to_duration_string( duration );" & vbLf & _
"                        line.USER_FORMAT.days        := duration / DAY;" & vbLf & _
"" & vbLf & _
"                        if sz = 0 then" & vbLf & _
"                            line := folder_used ? line : remove( line, rank.USER_FORMAT.folder );" & vbLf & _
"                            text := text & {line};" & vbLf & _
"                        end if;" & vbLf & _
"" & vbLf & _
"                        for idx_op : NUMERIC is 1 to sz loop" & vbLf & _
"                            sz       := size( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.exps );" & vbLf & _
"                            rate     := (sz = 0) ? (rate + steps(3)) : rate;" & vbLf & _
"                            steps(4) := (sz = 0) ? NULL : (steps(3) / sz);" & vbLf & _
"                            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                            line.USER_FORMAT.job_name     := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.job_name     );" & vbLf & _
"                            line.USER_FORMAT.job_result   := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.job_result   );" & vbLf & _
"                            line.USER_FORMAT.first_fatal  := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.first_fatal  );" & vbLf & _
"                            line.USER_FORMAT.project_name := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.project_name );" & vbLf & _
"                            line.USER_FORMAT.exp_count    := sz;" & vbLf & _
"" & vbLf & _
"                            if sz = 0 then" & vbLf & _
"                                line := folder_used ? line : remove( line, rank.USER_FORMAT.folder );" & vbLf & _
"                                text := text & {line};" & vbLf & _
"                                line := repeat( {""""}, rank.USER_FORMAT );" & vbLf & _
"                            end if;" & vbLf & _
"" & vbLf & _
"                            for idx_exp : NUMERIC is 1 to sz loop" & vbLf & _
"                                rate := rate + steps(4);" & vbLf & _
"                                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                                line.USER_FORMAT.data_file := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.data_file );" & vbLf & _
"                                line.USER_FORMAT.exp_file  := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.exp_file  );" & vbLf & _
"                                line.USER_FORMAT.folder    := to_string( data_by_user(idx_user).USER_DATA.owners(idx_owner).OWNER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.folder    );" & vbLf & _
"" & vbLf & _
"                                line := folder_used ? line : remove( line, rank.USER_FORMAT.folder );" & vbLf & _
"                                text := text & {line};" & vbLf & _
"                                line := repeat( {""""}, rank.USER_FORMAT );" & vbLf & _
"                            end loop;" & vbLf & _
"                        end loop;" & vbLf & _
"                    end loop;" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"            when ""Project Name"" =>" & vbLf & _
"                text := text & {{""Project Name"",""Operation Time"",""Operation Time/day"",""Jobs"",""Experiments""}};" & vbLf & _
"" & vbLf & _
"                sz       := size( order );" & vbLf & _
"                rate     := (sz = 0) ? (rate + 10) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (10 / sz);" & vbLf & _
"" & vbLf & _
"                for idx_proj in order loop" & vbLf & _
"                    rate := rate + steps(1);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                    line := repeat( {""""}, rank.PROJECT_SUMMARY_FORMAT );" & vbLf & _
"                    line.PROJECT_SUMMARY_FORMAT.project_name :=           to_string( data_by_proj(idx_proj).PROJECT_DATA.project_name, ""(No Project Name)"" );" & vbLf & _
"                    line.PROJECT_SUMMARY_FORMAT.op_time      :=  to_duration_string( data_by_proj(idx_proj).PROJECT_DATA.op_time );" & vbLf & _
"                    line.PROJECT_SUMMARY_FORMAT.op_days      :=                      data_by_proj(idx_proj).PROJECT_DATA.op_time / DAY;" & vbLf & _
"                    line.PROJECT_SUMMARY_FORMAT.op_count     :=                size( data_by_proj(idx_proj).PROJECT_DATA.ops );" & vbLf & _
"                    line.PROJECT_SUMMARY_FORMAT.exp_count    :=                      data_by_proj(idx_proj).PROJECT_DATA.exp_count;" & vbLf & _
"                    text := text & {line};" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"                rate     := (sz = 0) ? (rate + 90) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (90 / sz);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                for idx_proj in order loop" & vbLf & _
"                    sz       := size( data_by_proj(idx_proj).PROJECT_DATA.ops );" & vbLf & _
"                    rate     := (sz = 0 or else idx_proj not in selected) ? (rate + steps(1)) : rate;" & vbLf & _
"                    steps(2) := (sz = 0) ? NULL : (steps(1) / sz);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"                    next when idx_proj not in selected;" & vbLf & _
"" & vbLf & _
"                    line := {""Operation Began"",""Operation End"",""Duration"",""Duration/day"",""Username"",""Job Name"",""Experiment Name"",""Project Name"",""Folder"",""Filename""};" & vbLf & _
"                    line := folder_used ? line : remove( line, rank.PROJECT_FORMAT.folder );" & vbLf & _
"                    text := text & {""""," & vbLf & _
"                                    ""* Breakdown of project '"" & to_string( data_by_proj(idx_proj).PROJECT_DATA.project_name, ""(No Project Name)"" ) & ""'""," & vbLf & _
"                                    line};" & vbLf & _
"" & vbLf & _
"                    for idx_op : NUMERIC is 1 to sz loop" & vbLf & _
"                        sz       := size( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.exps );" & vbLf & _
"                        rate     := (sz = 0) ? (rate + steps(2)) : rate;" & vbLf & _
"                        steps(3) := (sz = 0) ? NULL : (steps(2) / sz);" & vbLf & _
"                        return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                        duration := data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.stop -" & vbLf & _
"                                    data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.start;" & vbLf & _
"" & vbLf & _
"                        line := new PROJECT_FORMAT;" & vbLf & _
"                        line.PROJECT_FORMAT.op_start     := to_date_time_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.start );" & vbLf & _
"                        line.PROJECT_FORMAT.op_stop      := to_date_time_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.stop  );" & vbLf & _
"                        line.PROJECT_FORMAT.username     :=                      data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.username;" & vbLf & _
"                        line.PROJECT_FORMAT.job_name     :=           to_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.job_name );" & vbLf & _
"                        line.PROJECT_FORMAT.project_name :=           to_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.project_name );" & vbLf & _
"                        line.PROJECT_FORMAT.duration     := to_duration_string( duration );" & vbLf & _
"                        line.PROJECT_FORMAT.days         := duration / DAY;" & vbLf & _
"" & vbLf & _
"                        for idx_exp : NUMERIC is 1 to sz loop" & vbLf & _
"                            rate := rate + steps(3);" & vbLf & _
"                            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                            line.PROJECT_FORMAT.data_file := to_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.data_file );" & vbLf & _
"                            line.PROJECT_FORMAT.exp_file  := to_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.exp_file  );" & vbLf & _
"                            line.PROJECT_FORMAT.folder    := to_string( data_by_proj(idx_proj).PROJECT_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.folder    );" & vbLf & _
"" & vbLf & _
"                            line := folder_used ? line : remove( line, rank.PROJECT_FORMAT.folder );" & vbLf & _
"                            text := text & {line};" & vbLf & _
"                            line := repeat( {""""}, rank.PROJECT_FORMAT );" & vbLf & _
"                        end loop;" & vbLf & _
"                    end loop;" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"            when ""Folder Name"" =>" & vbLf & _
"                text := text & {{""Folder Name"",""Operation Time"",""Operation Time/day"",""Jobs"",""Experiments""}};" & vbLf & _
"" & vbLf & _
"                sz       := size( order );" & vbLf & _
"                rate     := (sz = 0) ? (rate + 10) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (10 / sz);" & vbLf & _
"" & vbLf & _
"                for idx_folder in order loop" & vbLf & _
"                    rate := rate + steps(1);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                    line := repeat( {""""}, rank.FOLDER_SUMMARY_FORMAT );" & vbLf & _
"                    line.FOLDER_SUMMARY_FORMAT.folder    :=           to_string( data_by_folder(idx_folder).FOLDER_DATA.folder, ""(No Folder Name)"" );" & vbLf & _
"                    line.FOLDER_SUMMARY_FORMAT.op_time   :=  to_duration_string( data_by_folder(idx_folder).FOLDER_DATA.op_time );" & vbLf & _
"                    line.FOLDER_SUMMARY_FORMAT.op_days   :=                      data_by_folder(idx_folder).FOLDER_DATA.op_time / DAY;" & vbLf & _
"                    line.FOLDER_SUMMARY_FORMAT.op_count  :=                size( data_by_folder(idx_folder).FOLDER_DATA.ops );" & vbLf & _
"                    line.FOLDER_SUMMARY_FORMAT.exp_count :=                      data_by_folder(idx_folder).FOLDER_DATA.exp_count;" & vbLf & _
"                    text := text & {line};" & vbLf & _
"                end loop;" & vbLf & _
"" & vbLf & _
"                rate     := (sz = 0) ? (rate + 90) : rate;" & vbLf & _
"                steps(1) := (sz = 0) ? NULL : (90 / sz);" & vbLf & _
"                return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                for idx_folder in order loop" & vbLf & _
"                    sz       := size( data_by_folder(idx_folder).FOLDER_DATA.ops );" & vbLf & _
"                    rate     := (sz = 0 or else idx_folder not in selected) ? (rate + steps(1)) : rate;" & vbLf & _
"                    steps(2) := (sz = 0) ? NULL : (steps(1) / sz);" & vbLf & _
"                    return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"                    next when idx_folder not in selected;" & vbLf & _
"" & vbLf & _
"                    line := {""Operation Began"",""Operation End"",""Duration"",""Duration/day"",""Username"",""Job Name"",""Experiment Name"",""Project Name"",""Folder"",""Filename""};" & vbLf & _
"                    line := folder_used ? line : remove( line, rank.FOLDER_FORMAT.folder );" & vbLf & _
"                    text := text & {""""," & vbLf & _
"                                    ""* Breakdown of folder '"" & to_string( data_by_folder(idx_folder).FOLDER_DATA.folder, ""(No Folder Name)"" ) & ""'""," & vbLf & _
"                                    line};" & vbLf & _
"" & vbLf & _
"                    for idx_op : NUMERIC is 1 to sz loop" & vbLf & _
"                        sz       := size( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.exps );" & vbLf & _
"                        rate     := (sz = 0) ? (rate + steps(2)) : rate;" & vbLf & _
"                        steps(3) := (sz = 0) ? NULL : (steps(2) / sz);" & vbLf & _
"                        return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                        duration := data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.stop -" & vbLf & _
"                                    data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.start;" & vbLf & _
"" & vbLf & _
"                        line := new FOLDER_FORMAT;" & vbLf & _
"                        line.FOLDER_FORMAT.op_start     := to_date_time_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.start );" & vbLf & _
"                        line.FOLDER_FORMAT.op_stop      := to_date_time_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.stop  );" & vbLf & _
"                        line.FOLDER_FORMAT.username     :=                      data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.username;" & vbLf & _
"                        line.FOLDER_FORMAT.job_name     :=           to_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.job_name );" & vbLf & _
"                        line.FOLDER_FORMAT.project_name :=           to_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.project_name );" & vbLf & _
"                        line.FOLDER_FORMAT.duration     := to_duration_string( duration );" & vbLf & _
"                        line.FOLDER_FORMAT.days         := duration / DAY;" & vbLf & _
"" & vbLf & _
"                        for idx_exp : NUMERIC is 1 to sz loop" & vbLf & _
"                            rate := rate + steps(3);" & vbLf & _
"                            return when seconds - prev > UPDATE_INTERVAL and then handle_state( ""is aborted"", rate, @prev );" & vbLf & _
"" & vbLf & _
"                            line.FOLDER_FORMAT.data_file := to_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.data_file );" & vbLf & _
"                            line.FOLDER_FORMAT.exp_file  := to_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.exp_file  );" & vbLf & _
"                            line.FOLDER_FORMAT.folder    := to_string( data_by_folder(idx_folder).FOLDER_DATA.ops(idx_op).OP_DATA.exps(idx_exp).EXP_DATA.folder    );" & vbLf & _
"" & vbLf & _
"                            line := folder_used ? line : remove( line, rank.FOLDER_FORMAT.folder );" & vbLf & _
"                            text := text & {line};" & vbLf & _
"                            line := repeat( {""""}, rank.FOLDER_FORMAT );" & vbLf & _
"                        end loop;" & vbLf & _
"                    end loop;" & vbLf & _
"                end loop;" & vbLf & _
"        end case;" & vbLf & _
"" & vbLf & _
"        -- save CSV flie" & vbLf & _
"        result := save_csv_file( csv_file, !text );" & vbLf & _
"        if type result = [STRING] then" & vbLf & _
"            msg := ""Saved "" & pretty_name & "" successfully."";" & vbLf & _
"            sd {msg, csv_file};" & vbLf & _
"            log_message( msg, MSG_INFO );" & vbLf & _
"        else" & vbLf & _
"            msg := ""Failed to save "" & pretty_name & ""."";" & vbLf & _
"            result := (type result = [SET] and then size( result ) = 1) ? result(1) : result;" & vbLf & _
"            sd {msg, result};" & vbLf & _
"            log_message( msg &  "" : "" & [STRING](result), MSG_ERROR );" & vbLf & _
"        end if;" & vbLf & _
"        show_message( ""Please click [Save] icon to save current machine time data into CSV file."", MSG_INFO, ""enable save"" );" & vbLf & _
"    end HANDLE_SAVE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    function HANDLE_SET_VALUE( pname : STRING, pvalue : WILDCARD ) is" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        define CHECK_CONFIGURATION is" & vbLf & _
"            if conf = {start_date, stop_date, target_field} then" & vbLf & _
"                show_message( ""Please click [Save] icon to save current machine time data into CSV file."", MSG_INFO, ""enable save"" );" & vbLf & _
"            else" & vbLf & _
"                show_message( ""Please check [Start Date]/[Stop Date] and click [Calculate] button."", MSG_INFO, ""enable calculate"" );" & vbLf & _
"            end if;" & vbLf & _
"        end CHECK_CONFIGURATION;" & vbLf & _
"" & vbLf & _
"        -----------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"        var" & vbLf & _
"            time       : UNIVERSAL," & vbLf & _
"            start_time : UNIVERSAL," & vbLf & _
"            stop_time  : UNIVERSAL;" & vbLf & _
"" & vbLf & _
"        case pname is" & vbLf & _
"            when ""spec"" =>" & vbLf & _
"                set_value( ""MACHINE_TIME_CALCULATOR.SPECTROMETER"", pvalue );" & vbLf & _
"" & vbLf & _
"            when ""user"" =>" & vbLf & _
"                set_value( ""MACHINE_TIME_CALCULATOR.USERNAME"", pvalue );" & vbLf & _
"" & vbLf & _
"            when ""start date"" | ""stop date"" =>" & vbLf & _
"                time := to_date_time( pvalue );" & vbLf & _
"                time := (to_date_string(time) = pvalue) ? time : NULL;" & vbLf & _
"                start_time := (pname = ""start date"" and then type time = [UNIVERSAL]) ? time :  to_date_time( start_date );" & vbLf & _
"                stop_time  := (pname = ""stop date""  and then type time = [UNIVERSAL]) ? time :  to_date_time(  stop_date );" & vbLf & _
"                if start_time > stop_time then" & vbLf & _
"                    @start_time swap @stop_time;" & vbLf & _
"                end if;" & vbLf & _
"                start_date := to_date_string( start_time );" & vbLf & _
"                stop_date  := to_date_string(  stop_time );" & vbLf & _
"                check_configuration;" & vbLf & _
"" & vbLf & _
"            when ""target"" =>" & vbLf & _
"                set_value( ""MACHINE_TIME_CALCULATOR.TARGET"", pvalue );" & vbLf & _
"                target_field := pvalue;" & vbLf & _
"                check_configuration;" & vbLf & _
"" & vbLf & _
"            when ""group"" =>" & vbLf & _
"                set_value( ""MACHINE_TIME_CALCULATOR.GROUP"", pvalue );" & vbLf & _
"                sw {def, {""lb_log"", {WM~SET_FEATURE, WM~NO_DRAW, TRUE}}," & vbLf & _
"                         {""tbl_data""} & (pvalue = ""Username"") ? table_by_user : ((pvalue = ""Project Name"") ? table_by_proj : table_by_folder)};" & vbLf & _
"        end case;" & vbLf & _
"        check_status;" & vbLf & _
"    end HANDLE_SET_VALUE;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define HANDLE_SHUTDOWN is" & vbLf & _
"        -- user is shutting down Delta from Master Console" & vbLf & _
"        shutdown := TRUE;" & vbLf & _
"        sw {WM~SEND_ADDRESS, {def,""fr_main""}, {WM~WINQUIT}};" & vbLf & _
"    end HANDLE_SHUTDOWN;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    define HANDLE_QUIT return BOOLEAN is" & vbLf & _
"        -- delete temporary files" & vbLf & _
"        for temp_file in temp_files loop" & vbLf & _
"            call( ""code_mgr"", {CM~DELETE_SOURCE, temp_file}, EVENT );" & vbLf & _
"        end loop;" & vbLf & _
"" & vbLf & _
"        if nc /= null_netcontext then" & vbLf & _
"            fs_disconnect( nc );" & vbLf & _
"        end if;" & vbLf & _
"        unregister_node( inst_availability, def );" & vbLf & _
"        return TRUE;" & vbLf & _
"    end HANDLE_QUIT;" & vbLf & _
"" & vbLf & _
"    ---------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"    initialize;" & vbLf & _
"" & vbLf & _
"    register ""action""                 is handle_action;" & vbLf & _
"    register ""calendar""               is handle_calendar;" & vbLf & _
"    register ""calendar month""         is handle_calendar_month;" & vbLf & _
"    register ""calendar select""        is handle_calendar_select;" & vbLf & _
"    register ""instrument_available""   is handle_update_specs;" & vbLf & _
"    register ""instrument_unavailable"" is handle_update_specs;" & vbLf & _
"    register ""save""                   is handle_save;" & vbLf & _
"    register ""set value""              is handle_set_value;" & vbLf & _
"    register ""shutdown""               is handle_shutdown;" & vbLf & _
"    register ""quit""                   is handle_quit;" & vbLf & _
"" & vbLf & _
"    loop" & vbLf & _
"        exit when execute( receive( def, EVENT ) );" & vbLf & _
"    end loop;" & vbLf & _
"" & vbLf & _
"    consume( EVENT, def );" & vbLf & _
"    if shutdown then" & vbLf & _
"        send( ""shutdown"", {def}, MESSAGE );" & vbLf & _
"    else" & vbLf & _
"        send( !!""$$master_console"", {""QUIT_HANDSHAKE"",def,FALSE}, EVENT );" & vbLf & _
"    end if;" & vbLf & _
"end MACHINE_TIME_CALCULATOR;" & vbLf & _
""
End Function

' calendar_widget.t -------------------------------------------------------------
Function get_calendar_widget_t()
  get_calendar_widget_t = _
"-------------------------------------------------------------------------------" & vbLf & _
"--                                                                           --" & vbLf & _
"--                           Percival Source Code                            --" & vbLf & _
"--                Delta NMR Processing and Control Interface                 --" & vbLf & _
"--                                                                           --" & vbLf & _
"--                     Copyright (c) 2007 JEOL USA, Inc.                     --" & vbLf & _
"--                           All Rights Reserved                             --" & vbLf & _
"--                                                                           --" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"include ""callsend"";" & vbLf & _
"include ""color_schemes"";" & vbLf & _
"include ""dates"";" & vbLf & _
"include ""locale"";" & vbLf & _
"" & vbLf & _
"var" & vbLf & _
"    cdef           : SET     := {}," & vbLf & _
"    calendar_day   : NUMERIC," & vbLf & _
"    calendar_month : NUMERIC," & vbLf & _
"    calendar_year  : NUMERIC;" & vbLf & _
"" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"define CALENDAR_TABLE_SETUP return SET is" & vbLf & _
"    return {" & vbLf & _
"            {WM~SET_VALUE, WM~SET_COLUMN_HEADING, day_letter}," & vbLf & _
"            {WM~SET_VALUE, WM~SET_COLUMN_TYPE, repeat( {""string""}, size( day_abbrev ) )}," & vbLf & _
"            {WM~SET_VALUE, WM~SET_COLUMN_ALIGNMENT, repeat( {""center""}, size( day_abbrev ) )}" & vbLf & _
"           };" & vbLf & _
"end CALENDAR_TABLE_SETUP;" & vbLf & _
"" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function CREATE_CALENDAR_WINDOW( win_id : STRING, parent_widget : STRING ) return STRING is" & vbLf & _
"    var" & vbLf & _
"        fsize  : NUMBER := 11," & vbLf & _
"        cal_id : STRING := unique;" & vbLf & _
"" & vbLf & _
"    cdef(cal_id) := win_id;" & vbLf & _
"" & vbLf & _
"    if ""eng-"" not in get_locale then" & vbLf & _
"        fsize := 8;" & vbLf & _
"    end if;" & vbLf & _
"" & vbLf & _
"    sw {win_id," & vbLf & _
"        {WM~FRAME, ""FRM"" & cal_id, parent_widget," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, -1[mm],-1[mm],0[mm],0[mm]}," & vbLf & _
"            {WM~SET_BACKGROUND, ""grey65""}," & vbLf & _
"            {WM~SET_FEATURE, WM~BORDER, FALSE}," & vbLf & _
"            {WM~SET_FEATURE, WM~NO_DRAW, TRUE}" & vbLf & _
"        }," & vbLf & _
"        {WM~LABEL, ""MONTH"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, 0[mm],-11.5[mm],0[mm],0.5[mm]}," & vbLf & _
"            {WM~SET_VALUE, """"}," & vbLf & _
"            {WM~SET_VALUE, WM~GRAPHICS_OBJECT, {{WM~SET_COLOR, ""grey50""},{WM~DRAW_LINE, {0,50},{100,50}}}}," & vbLf & _
"            {WM~SET_FONT, 12}," & vbLf & _
"            {WM~SET_FOREGROUND, ""white""}," & vbLf & _
"            {WM~SET_BACKGROUND, color_title}," & vbLf & _
"            {WM~BINDING, {WM~MOUSE,WM~WILDCARD,WM~PRESS}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0, {""calendar month"",cal_id,0}}" & vbLf & _
"        }," & vbLf & _
"        {WM~PUSH_BUTTON, ""PREV"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, 0,-6[mm],-6[mm],0.5[mm]}," & vbLf & _
"            {WM~SET_VALUE, ""&#x2190;""}," & vbLf & _
"            {WM~SET_INFO, #""View Previous Month""}," & vbLf & _
"            {WM~SET_FONT, 14}," & vbLf & _
"            {WM~SET_HIGHLIGHT, {220,220,255}}," & vbLf & _
"            {WM~SET_FEATURE, WM~BORDER, ""GLOW""}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0, {""calendar month"",cal_id,-1}}" & vbLf & _
"        }," & vbLf & _
"        {WM~PUSH_BUTTON, ""NEXT"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, -6[mm],-6[mm],0,0.5[mm]}," & vbLf & _
"            {WM~SET_VALUE, ""&#x2192;""}," & vbLf & _
"            {WM~SET_INFO, #""View Next Month""}," & vbLf & _
"            {WM~SET_FONT, 14}," & vbLf & _
"            {WM~SET_HIGHLIGHT, {220,220,255}}," & vbLf & _
"            {WM~SET_FEATURE, WM~BORDER, ""GLOW""}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0, {""calendar month"",cal_id,1}}" & vbLf & _
"        }," & vbLf & _
"        {WM~PUSH_BUTTON, ""PREVYEAR"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, 0,-11.5[mm],-6[mm],6[mm]}," & vbLf & _
"            {WM~SET_VALUE, ""&#x21A9;""}," & vbLf & _
"            {WM~SET_INFO, #""View Previous Year""}," & vbLf & _
"            {WM~SET_FONT, 14}," & vbLf & _
"            {WM~SET_HIGHLIGHT, {220,220,255}}," & vbLf & _
"            {WM~SET_FEATURE, WM~BORDER, ""GLOW""}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0, {""calendar month"",cal_id,-12}}" & vbLf & _
"        }," & vbLf & _
"        {WM~PUSH_BUTTON, ""NEXTYEAR"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, -6[mm],-11.5[mm],0,6[mm]}," & vbLf & _
"            {WM~SET_VALUE, ""&#x21AA;""}," & vbLf & _
"            {WM~SET_INFO, #""View Next Year""}," & vbLf & _
"            {WM~SET_FONT, 14}," & vbLf & _
"            {WM~SET_HIGHLIGHT, {220,220,255}}," & vbLf & _
"            {WM~SET_FEATURE, WM~BORDER, ""GLOW""}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_STATE,WM~BUTTON_ON}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0, {""calendar month"",cal_id,12}}" & vbLf & _
"        }," & vbLf & _
"        {WM~TABLE, ""CAL"" & cal_id, ""FRM"" & cal_id," & vbLf & _
"            {WM~LEFT_BOTTOM_RIGHT_TOP, 2[mm],2[mm],2[mm],12.5[mm]}," & vbLf & _
"            {WM~SET_FONT, fsize}," & vbLf & _
"            {WM~SET_HIGHLIGHT, ""grey75""}," & vbLf & _
"            {WM~SET_FEATURE, WM~COLUMNAR, FALSE}," & vbLf & _
"            {WM~SET_FEATURE, WM~SEQUENCE, FALSE}," & vbLf & _
"            {WM~SET_FEATURE, WM~X_RULER, FALSE}," & vbLf & _
"            {WM~SET_FEATURE, WM~Y_RULER, FALSE}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_SELECTED,WM~CELL}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0," & vbLf & _
"                         {""calendar select"",cal_id,WM~WILDCARD}}," & vbLf & _
"            {WM~BINDING, {WM~RETURN_SELECTED,{}}," & vbLf & _
"                         WM~SEND_PERCIVAL, 0," & vbLf & _
"                         {""calendar select"",cal_id,WM~REGISTER,1}}" & vbLf & _
"        }" & vbLf & _
"      };" & vbLf & _
"" & vbLf & _
"    return cal_id;" & vbLf & _
"end CREATE_CALENDAR_WINDOW;" & vbLf & _
"" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function GET_CALENDAR_DATES( month : NUMERIC, year : NUMERIC ) return SET is" & vbLf & _
"    var" & vbLf & _
"        mvalid : BOOLEAN := type month = [NUMERIC] and then month in 1..12," & vbLf & _
"        yvalid : BOOLEAN := type year = [NUMERIC] and then year > 0," & vbLf & _
"        dat    : STRING," & vbLf & _
"        dow    : NUMERIC," & vbLf & _
"        d      : NUMERIC," & vbLf & _
"        w      : SET," & vbLf & _
"        cal    : SET;" & vbLf & _
"" & vbLf & _
"    if not mvalid or else not yvalid then" & vbLf & _
"        var" & vbLf & _
"            dt : STRING := date( seconds );" & vbLf & _
"" & vbLf & _
"        if not mvalid then" & vbLf & _
"            month := index( nth( dt, 4, 3), month_abbrev );" & vbLf & _
"        end if;" & vbLf & _
"        if not yvalid then" & vbLf & _
"            year := value( nth(dt, 8, 4) );" & vbLf & _
"        end if;" & vbLf & _
"    end if;" & vbLf & _
"" & vbLf & _
"    dow := day_of_week( 1, month, year );" & vbLf & _
"" & vbLf & _
"    cal := {repeat({""""}, dow - 1 ) & (1..(8 - dow))};" & vbLf & _
"    d   := 8 - dow + 1;" & vbLf & _
"    loop" & vbLf & _
"        exit when d + 6 > 28;" & vbLf & _
"        cal := cal & {d..(d + 6)};" & vbLf & _
"        d   := d + 7;" & vbLf & _
"    end loop;" & vbLf & _
"    loop" & vbLf & _
"        if d > 28 then" & vbLf & _
"            dat := to_time_stamp( d, month, year );" & vbLf & _
"            exit when dat /= date( date( dat ) );" & vbLf & _
"        end if;" & vbLf & _
"        w := w & {d};" & vbLf & _
"        if size( w ) = 7 then" & vbLf & _
"            cal := cal & {w};" & vbLf & _
"            w   := {};" & vbLf & _
"        end if;" & vbLf & _
"        d := d + 1;" & vbLf & _
"    end loop;" & vbLf & _
"    if size( w ) > 0 then" & vbLf & _
"        cal := cal & {w & repeat( {""""}, 7 - size( w ) )};" & vbLf & _
"    end if;" & vbLf & _
"    return cal;" & vbLf & _
"end GET_CALENDAR_DATES;" & vbLf & _
"" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"-- 'pos' must be in the form {WM~LEFT_BOTTOM_RIGHT_TOP, <left>, <bottom>, <right>, <top>}" & vbLf & _
"-- <top> or <bottom> MUST be given in [mm]." & vbLf & _
"-- <top> MUST be specified and <bottom> MUST be NULL, or visa versa" & vbLf & _
"--      - the calender with grow in the direction set to NULL" & vbLf & _
"-------------------------------------------------------------------------------" & vbLf & _
"function UPDATE_CALENDAR( cal_id : STRING, pos : SET ) is" & vbLf & _
"    var" & vbLf & _
"                     -- date( seconds )," & vbLf & _
"        d  : STRING  := to_time_stamp( calendar_day, calendar_month, calendar_year )," & vbLf & _
"        dy : NUMERIC," & vbLf & _
"        x  : NUMERIC," & vbLf & _
"        y  : NUMERIC," & vbLf & _
"        m  : SET;" & vbLf & _
"" & vbLf & _
"    m := get_calendar_dates( calendar_month, calendar_year );" & vbLf & _
"    if pos(3) = NULL and pos(5) /= NULL then" & vbLf & _
"        pos(3) := -pos(5) - 16.5[mm] - 5[mm] * size( m );" & vbLf & _
"    else" & vbLf & _
"        pos(5) := -pos(3) - 16.5[mm] - 5[mm] * size( m );" & vbLf & _
"    end if;" & vbLf & _
"    if calendar_year = value( nth( d, 8, 4 ) ) and then" & vbLf & _
"       calendar_month = index( nth( d, 4, 3), month_abbrev ) then" & vbLf & _
"        dy := value( nth( d, 1, 2) );" & vbLf & _
"        for w : NUMERIC is 1 to size( m ) loop" & vbLf & _
"            for v : NUMERIC is 1 to size( m(w) ) loop" & vbLf & _
"                if dy = m(w,v) then" & vbLf & _
"                    x := w;" & vbLf & _
"                    y := v;" & vbLf & _
"                    exit;" & vbLf & _
"                end if;" & vbLf & _
"            end loop;" & vbLf & _
"        end loop;" & vbLf & _
"    end if;" & vbLf & _
"    sw {cdef(cal_id)," & vbLf & _
"        {""FRM"" & cal_id,   pos, {WM~SET_FEATURE, WM~NO_DRAW, FALSE}}," & vbLf & _
"        {""MONTH"" & cal_id, {WM~SET_VALUE, {month_names(calendar_month), [STRING](calendar_year)}}}," & vbLf & _
"        {""CAL"" & cal_id,   {WM~SET_VALUE, m}," & vbLf & _
"                           {WM~SET_SELECTED}, {WM~SET_SELECTED, x, y}," & vbLf & _
"                           {WM~SET_REGISTER, 1, dy}} &" & vbLf & _
"                           calendar_table_setup};" & vbLf & _
"end UPDATE_CALENDAR;" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function HIDE_CALENDAR( cal_id : STRING ) is" & vbLf & _
"    sw {{cdef(cal_id),""FRM"" & cal_id}, {WM~SET_FEATURE, WM~NO_DRAW, TRUE}," & vbLf & _
"                                       {WM~LEFT_BOTTOM_RIGHT_TOP, -1[mm],-1[mm],0[mm],0[mm]}};" & vbLf & _
"end HIDE_CALENDAR;" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function SET_CALENDAR( cal_id : STRING, ts : NUMBER, pos : SET ) is" & vbLf & _
"    var" & vbLf & _
"        dt : STRING := date( ts );" & vbLf & _
"" & vbLf & _
"    calendar_day   := value( nth( dt, 1, 2) );" & vbLf & _
"    calendar_month := index( nth( dt, 4, 3), month_abbrev );" & vbLf & _
"    calendar_year  := value( nth( dt, 8, 4) );" & vbLf & _
"" & vbLf & _
"    update_calendar( cal_id, pos );" & vbLf & _
"end SET_CALENDAR;" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function SET_CALENDAR_MONTH( cal_id : STRING, months : NUMERIC, pos : SET ) is" & vbLf & _
"    if months = 0 then" & vbLf & _
"        set_calendar( cal_id, seconds, pos );" & vbLf & _
"        return;" & vbLf & _
"" & vbLf & _
"    elsif months > 0 then" & vbLf & _
"        calendar_month := calendar_month - 1 + months;" & vbLf & _
"        calendar_year  := calendar_year + calendar_month div 12;" & vbLf & _
"" & vbLf & _
"    else" & vbLf & _
"        if months mod 12 = 0 then" & vbLf & _
"            calendar_year := calendar_year + months div 12;" & vbLf & _
"        elsif (abs months mod 12) >= calendar_month then" & vbLf & _
"            calendar_year := calendar_year + months div 12 - 1;" & vbLf & _
"        end if;" & vbLf & _
"        calendar_month := (calendar_month - 1) + (12 + months) mod 12;" & vbLf & _
"    end if;" & vbLf & _
"    calendar_month := calendar_month mod 12 + 1;" & vbLf & _
"    update_calendar( cal_id, pos );" & vbLf & _
"end SET_CALENDAR_MONTH;" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"function HANDLE_CALENDAR_MONTH( cal_id : STRING, months : NUMERIC );" & vbLf & _
"" & vbLf & _
"function HANDLE_CALENDAR_SELECT( cal_id : STRING, day : NUMERIC );" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"--Be sure to register these routines for your event loop with the following lines" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
"" & vbLf & _
"--register ""calendar month""  is handle_calendar_month;" & vbLf & _
"--register ""calendar select"" is handle_calendar_select;" & vbLf & _
"" & vbLf & _
"----------------------------------------------------------------------------" & vbLf & _
""
End Function
